/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		header file
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#ifndef _N80_H_
#define _N80_H_
#ifdef __linux__
#ifdef _ZAURUS_
#include <linux/fb.h>
#endif
#include <sys/time.h>
#include <time.h>
#include <pthread.h>
#include <SDL2/SDL.h>
#include <sys/utsname.h>
#include <alsa/asoundlib.h>
#include <stdarg.h>
#include <linux/limits.h>
#endif
#include <stdio.h>
#include "n80types.h"
#include "../z80/z80.h"
/**--------------------------------------------------------------------------
 **	define key codes
 **------------------------------------------------------------------------*/
#define VK_TAB		(0x0F)
#define VK_ESCAPE	(0x01)
#define VK_BACK		(0x0E)
#define VK_INSERT	(0x6E)
#define VK_DELETE	(0x6F)
#define VK_RETURN	(0x1C)
#define VK_SPACE	(0x39)
#define VK_UP		(0x67)
#define VK_DOWN		(0x6C)
#define VK_LEFT		(0x69)
#define VK_RIGHT	(0x6A)
#define VK_F6		(0x40)
#define VK_F7		(0x41)
#define VK_F8		(0x42)
#define VK_F9		(0x43)
#define VK_F10		(0x44)
#define VK_F12		(0x58)
#define VK_END		(0x6B)
/**--------------------------------------------------------------------------
 **		PROTOTYPE DECLARATIONs
 **------------------------------------------------------------------------*/
struct tagCFGITEM;
typedef struct tagCFGITEM	CFGITEM, *LPCFGITEM;
struct tagCFGPAGE;
typedef struct tagCFGPAGE	CFGPAGE, *LPCFGPAGE;
struct CONFIG_tag;
typedef struct CONFIG_tag	CONFIG, *LPCONFIG;
struct CHOOSER_tag;
typedef struct CHOOSER_tag	CHOOSER, *LPCHOOSER;
struct INI_tag;
typedef struct INI_tag		INI, *PINI;
struct BDF_tag;
typedef struct BDF_tag		BDF, *PBDF;
struct BDF_BMP_tag;
typedef struct BDF_BMP_tag	BDFBMP, *PBDFBMP;
struct LX_alsa_devs_tag;
typedef struct LX_alsa_devs_tag	ALSADEVs, *PALSADEVs;
struct N80LX_tag;
typedef struct N80LX_tag	N80LX, *LPN80LX;
/**--------------------------------------------------------------------------
 **	declare structure types
 **------------------------------------------------------------------------*/
typedef struct LX_alsa_dev_tag {
	char	dev[80];
	char	name[80];
}	ALSADEV, *PALSADEV;

struct LX_alsa_devs_tag {
	int			rows_per_page;
	int			pages;
	int			cur_page;
	int			rows;
	int			cur_row;
	int			rack_page;
	int			rack_page_rows;
	int			selected;
	int			cnt;
	PALSADEV	list;
	int			(*rows_in_page)(PALSADEVs, int);
	int			(*page_start_index)(PALSADEVs);
	int			(*page_end_index)(PALSADEVs);
	int			(*row_to_index)(PALSADEVs);
	void		(*show_page)(PALSADEVs);
	bool		(*pgdn)(PALSADEVs);
	bool		(*pgup)(PALSADEVs);
	bool		(*down)(PALSADEVs);
	bool		(*up)(PALSADEVs);
};

struct BDF_BMP_tag {
	uint16_t	jis;
	BYTE		bitmap[2][16];
};
struct BDF_tag {
	char		path[PATH_MAX];
	int			cnt;
	PBDFBMP		bitmap;
	void		(*dispose)(PBDF);
	bool		(*load)(PBDF);
	PBDFBMP		(*find)(PBDF, uint16_t);
	uint16_t	(*sjis2jis)(uint16_t);
	uint16_t	(*jis2sjis)(uint16_t);
};

typedef struct tagKEYEVT {
    USHORT		key;
	BOOL		SHIFT;
	BOOL		ALT;
	BOOL		CTL;
    char		ascii;
}   KEYEVTG, *LPKEYEVT;

struct tagCFGITEM {
	int			id;
	int			type;
	char*		name;
	char*		hint;
	int			value;
	int			(*controller)(LPCFGITEM, int, LPVOID);
	LPVOID		arg;
	LPCFGPAGE	parent;
};

struct tagCFGPAGE {
	LPCONFIG	parent;
	int			page;
	LPVOID		arg;
	LPCFGITEM	items;
	int			item_cnt;
	int			cur_idx;
	LPCFGITEM	(*get_item)(LPCFGPAGE);
	int			(*handler)(LPCFGPAGE, LPKEYEVT);
	void		(*prepare)(LPCFGPAGE, LPVOID arg);
	int			(*call)(LPCFGPAGE, int, LPVOID);
	void		(*reflesh)(LPCFGPAGE);
	int			jmp_page;
	LPVOID		jmp_arg;
};

struct CONFIG_tag {
	int			fd;
	KEYEVTG		kv;
	CFGPAGE		pages[2];
	int			page;
	int			maxpages;
	void		(*init)(LPCONFIG);
	void		(*open)(LPCONFIG);
	void		(*init_page)(LPCONFIG, int page, LPVOID arg);
	void		(*evt_pump)(LPCONFIG);
	BOOL		(*dispatch)(LPCONFIG, LPKEYEVT);
	void		(*drawitem)(LPCFGITEM item, int line, BOOL selected);
	void		(*show)(LPCONFIG);
	char		(*to_ascii)(LPKEYEVT kv, SDL_Event*);
};

struct CHOOSER_tag {
	char		dirpath[PATH_MAX];
	void		(*open)(LPCHOOSER, int slot_no, int unit_no);
	char*		(*utf8to932)(const char *utf);
	BOOL 		(*read_d88file)(const char *, int slot_no, int unit_no);
	BOOL		(*read_t88file)(const char *);
	BOOL		(*read_cmtfile)(const char *);
	BOOL 		(*read_n80file)(const char *);
};

struct INI_tag {
	void	(*init)(PINI);
	BOOL	(*flush)(PINI);
	BOOL	(*load)(PINI);
	void	(*dispose)(PINI);
};

typedef struct _callback_info	callback_info;
typedef int	(*callback)(callback_info *param);
struct _callback_info {
	SDL_Event	event;
	callback	handler;
	callback	subhandler;
	LPVOID		param;
};

struct N80LX_tag {
	#ifdef __linux__
	SDL_Window*					window;						// SDL2 Window
	SDL_Renderer*				renderer;					// SDL2 Renderer
	SDL_Texture*				texture;					// SDL2 Texture
	Uint32						pipe_evt;					// pipe event for SDL_Event
	LPDWORD						video;						// texture pixcels
	ALSADEVs					snddevs;					// alsa pcm playback devices
	char						save_cmt_dir[511-17];
	char						save_cmt_name[17];
	bool						cpu_no_wait;
	bool						show_sys_clock;
	DWORD						sys_Ts;
	double						sys_hz;						// system clock (Hz)
	double						eff_hz;						// effective cpu clock (Hz)
	int							flips;						// draw screen per sec
	INI							ini;						// ini file
	CONFIG						config;						// config
	CHOOSER						chooser;					// image file chooser
	char						n80rom_filename[64];
	char						fdurom_filename[64];
	char						bdf_filename[64];			// bitmap kanji font (bdf)
	char						snd_dev_name[64];			// alsa device name
	#endif
	#ifdef _ZAURUS_
	USHORT						dot_pattern[8][256][16];	// fb graphics pattern data (must be alocated in iwmmx align)
	int							fbfd;						// fb descriptor
	LPUSHORT					video;						// fb pointer to mapped memory
	struct fb_var_screeninfo	vinfo;						// fb variable info
	struct fb_fix_screeninfo	finfo;						// fb fixed info
	#endif
	pthread_mutex_t				vram_sync;					// PC-8001 vram sync
	pthread_mutex_t				pause_sync;					// pause sync
	volatile DWORD 				request_cpu_signal;			// signal for z80 thread
	#ifdef __linux__
	struct utsname				utsname;					// os version
	int							tenkeyless;					// zaurus [Fn] key press status
	char						home[PATH_MAX];				// n80 home dir path
	#endif
	snd_pcm_t					*handle;					// alsa handle
	snd_pcm_hw_params_t			*params;					// alsa parameters
	int							sampling_rate;				// alsa pcm sampling rate
	int							period_frames;				// alsa period length in samples
	int							period_bytes;				// alsa period length in byte
	LPUSHORT					period_buffer;				// alsa local buffer
	int							period_usec;				// alsa period long in micor sec
	snd_pcm_sframes_t			ring_buffer_frames;			// alsa ring buffer length in samples
	double						frame_keisu;				// 48kHz(PCM) / 4MHz(CPU)
	int							phase_adjuster;				// wave phase adjust (PCG Sound)
	BDF							bdf;
	char						text[25][80];
	char						attr[25][80];
	bool						shift;
	bool						alt;
	bool						ctl;
	bool						ins;
	bool						del;
	BYTE						save_pcg;
};

enum N80COLOR {N80BLACK = 0, N80BLUE, N80RED, N80PERPLE, N80GREEN, N80CYAN, N80YELLOW, N80WHITE};
#ifdef _DEBUG_
#define XTRACE __trace
#else
#define XTRACE (void)
#endif//_DEBUG_
#endif//_N80_H_

