/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2021 H.Kanemoto
 **
 **		sound device selector
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"

#define export
/**--------------------------------------------------------------------------
 **	external value/function
 **------------------------------------------------------------------------*/
extern N80LX	n80lx;
extern void		draw_icon(LPDWORD video, int pitch, int x, int y);
extern void		draw_text(int line, LPBYTE attr, LPBYTE text);
extern void		draw_textk(int line, LPBYTE attr, LPBYTE text);
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static int page = 0, pages = 3;
static const char *const doc[][15] = {
	{
		"      キー操作",
		"    -----------------------------------------------------------------------",
		"        PC-8001 の KEY で 106 KEYBOARD にないものは、別の KEY に割り当",
		"        てられています：",
		"",
		"            PC-8001                    106 KEY",
		"            -------             -----------------------------",
		"            GRAPH               左[ALT]",
		"            カナ                右[ALT]",
		"            STOP                BackSpace",
		"            RESET               [F12]",
		"",
		"        また、「N80」独自の機能が 106 KEYBOARD に割り当てられています：",
		"",
		"",
	},{
		"            106 KEY                     機　能",
		"            -------             -----------------------------",
		"            [F6]                この画面",
		"            [F7]                設定画面",
		"            [F9]                N80/CMT/T88/D88 ﾌｧｲﾙ選択",
		"            [F11]               PCG-8100 ON/OFF",
		"            [END]               ｴﾐｭﾚｰﾀｰ終了",
		"",
		"            [SHIFT]+[F6]        ｼｽﾃﾑｽﾃｰﾀｽ表示(ﾄｸﾞﾙ式)",
		"            [SHIFT]+[F12]       全画面切り替え(ﾄｸﾞﾙ式)",
		"",
		"            [ALT]+[INS]+[1|2]   ﾌﾛｯﾋﾟｰﾃﾞｨｽｸ挿入 (slot 1 or 2)",
		"            [ALT]+[DEL]+[1|2]   ﾌﾛｯﾋﾟｰﾃﾞｨｽｸ排出 (slot 1 or 2)",
		"",
		"",
	},{
		"      サウンド出力デバイス (ALSA)",
		"    -----------------------------------------------------------------------",
		"        起動時に使用するPCM音声出力ﾃﾞﾊﾞｲｽの選択画面が表示されます。",
		"        一度選択されると次回以降も同じﾃﾞﾊﾞｲｽが使用されます。",
		"",
		"        ﾃﾞﾊﾞｲｽを変更したい時は、[F7]-設定画面から\"alsa sound device name\"",
		"        を消すか無効な名前にしてｴﾐｭﾚｰﾀを一旦終了して再起動します。",
		"        再度、選択画面が表示されます。",
		"",
		"",
		"      謝  辞",
		"    -----------------------------------------------------------------------",
		"        漢字ﾌｫﾝﾄに*Public Domain*として公開されている「東雲 ビットマップ",
		"        フォントファミリー」を使用させて頂いてます。",
		"        古川泰之さん、/efont/さん、ありがとうございます。",
	}
};
/**--------------------------------------------------------------------------
 **	internal functions
 **------------------------------------------------------------------------*/
static void show() {
	for (int r = 0; r < 25; ++r) {
		if (r == 7 || (r > 8 && r < 24))
			draw_textk(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
		else
			draw_text(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
	}
	draw_icon(n80lx.video, 640, 12, 16);
	SDL_UpdateTexture(n80lx.texture, NULL, n80lx.video, 640 * 4);
	SDL_RenderClear(n80lx.renderer);
	SDL_RenderCopy(n80lx.renderer, n80lx.texture, NULL, NULL);
	SDL_RenderPresent(n80lx.renderer);
}

static void showpage(int page) {
	if (page < 0 || page >= pages)
		return;
	for (int r = 0; r < 15; ++r) {
		char *line = n80lx.chooser.utf8to932(doc[page][r]);
		int len = strlen(line);
		for (int col = 1, cnt = 0; col < 79; ++col) {
			if (cnt < len)
				n80lx.text[r+9][col] = line[cnt++];
			else
				n80lx.text[r+9][col] = '\0';
		}
	}
	show();
}

static void pageup(void) {
	if (page > 0) {
		--page;
		showpage(page);
	} else
		return;
}

static void pagedn(void) {
	if ((page + 1) < pages) {
		++page;
		showpage(page);
	} else
		return;
}
/**--------------------------------------------------------------------------
 **	global function
 **------------------------------------------------------------------------*/
export void about(void) {
	for (int r = 0; r < 25; ++r) {
		for (int c = 0; c < 80; ++c) {
            BYTE atr = 0x23;
            if ((r == 4 || r == 5) && (c > 0 && c < 79))
                atr = 0x27;
            else if (r == 7 && (c > 0 && c < 79))
                atr = 0x26;
            else if ((r > 8 && r < 24) && (c > 0 && c < 79))
                atr = 0x07;
			n80lx.attr[r][c] = atr;
			n80lx.text[r][c] = 32;
		}
	}
	const char *title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for Linux";
	const char *title2 = "    Copyright 1993-2025 (C) H.Kanemoto, All Right Reserved.";
	const char *title3 = "Document";
	const char *title4 = "usage:  [ESC]...Quit, [PgUp/PgDn]...Page";
	const char *title5 = "hint: Explanation of N80";
	memcpy(&n80lx.text[1][2], title1, strlen(title1));
	memcpy(&n80lx.text[2][2], title2, strlen(title2));
	memcpy(&n80lx.text[4][2], title3, strlen(title3));
	memcpy(&n80lx.text[5][2], title4, strlen(title4));
	memcpy(&n80lx.text[7][2], title5, strlen(title5));

	page = 0;
	showpage(page);

	SDL_Event	event;
	for (int quit = 0; !quit;) {
		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_QUIT)
				quit = 1;
			else if (event.type == SDL_KEYDOWN) {
				if (event.key.keysym.sym == SDLK_ESCAPE) {
					if (event.key.repeat == 0) {
						quit = 1;	break;
					}
				} else if (event.key.keysym.sym == SDLK_PAGEDOWN) {
					pagedn();
				} else if (event.key.keysym.sym == SDLK_PAGEUP) {
					pageup();
				}
			}
		}
	}
}
