#==============================================================================
#
#	PC-8001 Device Emulator for Linux(x64) Ver 1.00
#	Copyright (C)1993-2025 H.Kanemoto
#
#	This is a free software. You may copy or distribute
#	it without any permissions. But there are NO WARRANTY
#	of use. Entire risk of use is with you.
#
#	External assembler functions for IBM-PC with gcc
#	last modified  Feb,7,2025.
#
#==============================================================================
			.file	"bltmmx.s"
			.data
			.align 32
__fnt_pat:	.skip	80
			.bss
			.section .note.GNU-stack,"",%progbits
			.text
			.intel_syntax	noprefix
			.align	8
#------------------------------------------------------------------------------
#		void	__cls_graphX (vram,lpitch)
#		rdi:	vram
#		esi:	lpitch
#------------------------------------------------------------------------------
			.global	__cls_graphX
			.type	__cls_graphX, @function
__cls_graphX:		
			push	rbp
			mov		rbp,rsp
			push	rbx
			####	recieve args
			xor		rbx,rbx
			mov		ebx,esi						# lpitch

			mov		rax,32
			mov		rdx,rdi						# save vram
			xor		rcx,rcx						# zero clear xmm0 reg
			movq	xmm0,rcx

			####	start drawing
			mov		rcx,200						# 200 rasters
_raster_loop:
			.rept	80							# 8pixel*80=640pixel/raster
			movdqa	[rdi],xmm0					# clear even raster
			movdqa	[rdi+16],xmm0
			movdqa	[rdi+rbx],xmm0				# clear odd raster
			movdqa	[rdi+rbx+16],xmm0
			add		rdi,rax						# ++vram
			.endr

			add		rdx,rbx						# vram next
			add		rdx,rbx
			mov		rdi,rdx
			dec		rcx
			jnz		_raster_loop

			####	end of draw
			pop		rbx
			pop		rbp
			ret
#------------------------------------------------------------------------------
#		macro	__adj_pat
#------------------------------------------------------------------------------
#		adjust font pattern according to attribute(pac) and save it
#			IN:		ah:attr al:pat(raw) rdi:*pat(cooked) rdx:*fnt(raw)
#			OUT:	al:pat(cooked)
#------------------------------------------------------------------------------
			.macro	__adj_pat
			test	ah,0x40						# check attr erase
			jz		1f
			xor		al,al
1:			test	ah,0x20						# check attr reverse
			jz		2f
			not		al
2:			nop
			.endm

			.macro	__adj_top					# top font raster
			test	ah,0x08						# check attr upper-line
			jz		1f
			mov		al,0xff
1:			__adj_pat
			.endm

			.macro	__adj_bot					# bottom font raster
			test	ah,0x80						# check attr under-line
			jz		1f
			mov		al,0xff
1:			__adj_pat
			.endm

			.macro	__bit_exp					# expand bit to double
			shl		bx,2
			sal		al,1
			jnc		1f
			or		bx,3
1:			nop
			.endm

			.macro	__fnt_exp					# expand 8pix narrow font pattern to 16pix wide
			.rept	8
			__bit_exp
			.endr
			.endm

			.macro	__fnt_sel					# get pointer to char font
			test	ah,0x10						# check graph char
			jz		1f
			lea		rdx,[rdi+2048+rcx]			# pointer to font (graph)
			jmp		3f
1:			test	r15b,1						# check use-pcg(hal) flag
			jz		2f
			lea		rdx,[rdi+4096+rcx]			# pointer to font (pcg)
			jmp		3f
2:			lea		rdx,[rdi+rcx]				# pointer to font (normal)
3:			nop
			.endm
#******************************************************************************
#			Graphic (80columns) T/G Mix
#******************************************************************************
			.macro	__adj_patN					# narrow font single raster
			__adj_pat
			__adj_saveN
			.endm
			.macro	__adj_topN					# narrow font top raster
			__adj_top
			__adj_saveN
			.endm
			.macro	__adj_botN					# narrow font bottom raster
			__adj_bot
			__adj_saveN
			.endm
			.macro	__adj_saveN	
			mov		BYTE PTR[rdi],al			# save cooked pat
			inc		rdi
			mov		al,BYTE PTR[rdx]			# get next raw pat
			inc		rdx
			.endm

			.macro	__adj_fnt8					# 8 rasters/char (25 lines)
			mov		al,BYTE PTR[rdx]
			inc		rdx
			__adj_topN
			.rept	6
			__adj_patN
			.endr
			__adj_botN
			.endm

			.macro	__adj_fnt10					# 10 rasters/char (20 lines)
			xor		al,al
			__adj_topN
			.rept	8
			__adj_patN
			.endr
			xor		al,al
			__adj_botN
			.endm
#------------------------------------------------------------------------------
#		macro	__fntrastN
#------------------------------------------------------------------------------
			.macro	__fntbit	chk,cp1
			sal		dl,1
			jnc		1f
			mov		eax,r13d					# retrieve color (even raster)
			mov		ebx,r14d					# retrieve color (odd raster)
			jmp		short 2f
1:			xor		eax,eax
			mov		ebx,eax
2:			movq	rdi,xmm5
			mov		DWORD PTR[rdi+\cp1],eax
			mov		DWORD PTR[rdi+rcx+\cp1],ebx
			.endm

			.macro	__fntdraw
			####	copy when bit=1
			__fntbit	0x80,0					# b7
			__fntbit	0x40,4					# b6
			__fntbit	0x20,8					# b5
			__fntbit	0x10,12					# b4
			__fntbit	8,16					# b3
			__fntbit	4,20					# b2
			__fntbit	2,24					# b1
			__fntbit	1,28					# b0
			.endm

			.macro	__fntrastN
			mov		dl,BYTE PTR[rsi]			# get font pattern
			movq	xmm3,rsi
			__fntdraw							# draw narrow font
			add		rdi,rcx
			add		rdi,rcx
			movq	xmm5,rdi					# save vram'
			movq	rsi,xmm3
			inc		rsi							# next pattern
			.endm
#------------------------------------------------------------------------------
#		void	__ovl_802X (txt[],pac[],font[],vram,pal[],lpitch,pcgon)
#------------------------------------------------------------------------------
#	---	args
#		rdi:	*txt
#		rsi:	*pac
#		rdx:	*font
#		rcx:	*vram
#		r8:		*pal
#		r9:		lpitch
#	---	work
#		xmm15	__adj_802x
#		xmm14	__ras_802x
#		xmm7	*fnt
#		xmm6	*vram
#		xmm5	*vram'
#		xmm4	*_pat
#		xmm3	tmp
#		xmm2	*txt
#		xmm1	*pac
#		xmm0	*pal
#		r13d	color #1
#		r14d	color #2
#		r15w	pcg_on_off
#------------------------------------------------------------------------------
			.text
			.global	__ovl_8025X,__ovl_8020X
			.type	__ovl_8025X, @function
			.type	__ovl_8020X, @function
__ovl_8025X:
			lea		rax,_adj_8025[rip]
			movq	xmm15,rax
			lea		rax,_ras_8025[rip]
			jmp		1f
__ovl_8020X:
			lea		rax,_adj_8020[rip]
			movq	xmm15,rax
			lea		rax,_ras_8020[rip]
1:			movq	xmm14,rax

			push	rbp
			mov		rbp,rsp
			push	rbx
			push	r13
			push	r14
			push	r15

			####	recieve args
			movq	xmm7,rdx					# *fnt
			movd	xmm6,rcx					# vram
			movd	xmm0,r8						# palette (font)
			mov		r15w,[rbp+16]				# pcguse
			lea		rax,__fnt_pat[rip]			# _pat
			movq	xmm4,rax
			mov		cx,80						# col counter
			####	draw columns loop
20:			push	rcx							# save col counter
			mov		al,BYTE PTR[rdi]			# ascii = *txt++
			inc		rdi
			movq	xmm2,rdi					# save *txt
			mov		ah,BYTE PTR[rsi]			# attr = *pac++
			inc		rsi
			movq	xmm1,rsi					# save *pac
			####	adjust font pointer
			movq	rdi,xmm7					# retrieve *fnt
			movzx	rcx,al						# index of fnt
			shl		rcx,3
			__fnt_sel							# get pointer to char-font
			####	adjust font bit pattern accrding to attribute(pac)
_adj_802x:	movq	rsi,xmm4					# retrieve *pat
			mov		rdi,rsi
			movq	rcx,xmm15					# retrieve address of _adj_802x
			jmp		rcx
_adj_8025:
			__adj_fnt8
			jmp		short 3f
_adj_8020:
			__adj_fnt10
			####	get font color
3:			movq	rdi,xmm0					# retrieve color palette
			and		ah,0x07						# pac.BRG
			mov		cl,ah
			movzx	rcx,cl
			shl		rcx,2
			mov		r13d,DWORD PTR[rdi+rcx]		# font color #1,#2
			mov		r14d,DWORD PTR[rdi+32+rcx]
			####	draw font/graph mix
			mov		rcx,r9						# retrieve lpitch
			movq	rdi,xmm6					# retrieve vram
			movq	xmm5,rdi					# save vram'
			movq	rax,xmm14					# retrieve address of _ras_802x
			jmp		rax
_ras_8020:
			__fntrastN
			__fntrastN
_ras_8025:
			.rept	8
			__fntrastN
			.endr

			movq	rbx,xmm6					# retrieve vram
			add		rbx,32						# next video addr
			movq	xmm6,rbx					# save vram
			movq	rsi,xmm1					# retrieve *pac
			movq	rdi,xmm2					# retrieve *txt
			pop		rcx							# retrieve col counter
			dec		cx
			jnz		20b

			####	end of draw
			pop		r15
			pop		r14
			pop		r13
			pop		rbx
			pop		rbp
			ret
#******************************************************************************
#			Graphic (40columns) T/G Mix
#******************************************************************************
			.macro	__adj_patW					# wide font single raster
			__adj_pat
			__adj_saveW
			.endm
			.macro	__adj_topW					# wide font single top raster
			__adj_top
			__adj_saveW
			.endm
			.macro	__adj_botW					# wide font single bottom raster
			__adj_bot
			__adj_saveW
			.endm
			.macro	__adj_saveW	
			xor		ebx,ebx						# expand narrow pattern to wide
			__fnt_exp
			mov		BYTE PTR[rdi],bh			# *pat2++ = pat (left)
			inc		rdi
			mov		BYTE PTR[rdi],bl			# *pat2++ = pat (right)
			inc		rdi
			mov		al,BYTE PTR[rdx]			# pat = *fnt++
			inc		rdx
			.endm

			.macro	__adj_fnt8W
			mov		al,BYTE PTR[rdx]
			inc		rdx
			__adj_topW
			.rept	6
			__adj_patW
			.endr
			__adj_botW
			.endm

			.macro	__adj_fnt10W
			xor		al,al
			__adj_topW
			.rept	8
			__adj_patW
			.endr
			xor		al,al
			__adj_botW
			.endm
#------------------------------------------------------------------------------
#		macro	__fntrastW
#------------------------------------------------------------------------------
			.macro	__fntrastW
			mov		dl,BYTE PTR[rsi]			# get font pattern
			movq	xmm3,rsi
			__fntdraw							# draw wide font [0] - left side
			add		rdi,32
			movq	xmm5,rdi					# save vram'
			mov		dl,BYTE PTR[rsi+1]
			__fntdraw							# draw wide font [1] - right side
			add		rdi,rcx
			add		rdi,rcx
			sub		rdi,32
			movq	xmm5,rdi					# save vram'
			movq	rsi,xmm3
			inc		rsi							# next pattern
			inc		rsi

			.endm
#------------------------------------------------------------------------------
#		void	__ovl_402X (txt[],pac[],font[],vram,pal[],lpitch,pcgon)
#------------------------------------------------------------------------------
#	---	args
#		rdi:	*txt
#		rsi:	*pac
#		rdx:	*font
#		rcx:	*vram
#		r8:		*pal
#		r9:		lpitch
#	---	work
#		xmm15	__adj_402x
#		xmm14	__ras_402x
#		xmm7	*fnt
#		xmm6	*vram
#		xmm5	*vram'
#		xmm4	*_pat
#		xmm3	tmp
#		xmm2	*txt
#		xmm1	*pac
#		xmm0	*pal
#		r13d	color #1
#		r14d	color #2
#		r15w	pcg_on_off
#------------------------------------------------------------------------------
			.text
			.global	__ovl_4025X,__ovl_4020X
			.type	__ovl_4025X, @function
			.type	__ovl_4020X, @function
__ovl_4025X:
			lea		rax,_adj_4025[rip]
			movq	xmm15,rax
			lea		rax,_ras_4025[rip]
			jmp		1f
__ovl_4020X:
			lea		rax,_adj_4020[rip]
			movq	xmm15,rax
			lea		rax,_ras_4020[rip]
1:			movq	xmm14,rax

			push	rbp
			mov		rbp,rsp
			push	rbx
			push	r13
			push	r14
			push	r15

			####	recieve args
			movq	xmm7,rdx					# *fnt
			movd	xmm6,rcx					# vram
			movd	xmm0,r8						# palette (font)
			mov		r15w,[rbp+16]				# pcguse
			lea		rax,__fnt_pat[rip]			# _pat
			movq	xmm4,rax
			mov		cx,40						# col counter
			####	draw columns loop
20:			push	rcx							# save col counter
			mov		al,BYTE PTR[rdi]			# ascii = *txt++
			inc		rdi
			inc		rdi
			movq	xmm2,rdi					# save *txt
			mov		ah,BYTE PTR[rsi]			# attr = *pac++
			inc		rsi
			inc		rsi
			movq	xmm1,rsi					# save *pac
			####	adjust font pointer
			movq	rdi,xmm7					# retrieve *fnt
			movzx	rcx,al						# index of fnt
			shl		rcx,3
			__fnt_sel							# get pointer to char-font
			####	adjust font bit pattern accrding to attribute(pac)
_adj_402x:	movq	rsi,xmm4					# retrieve *pat
			mov		rdi,rsi
			movq	rcx,xmm15					# retrieve address of _adj_802x
			jmp		rcx
_adj_4025:
			__adj_fnt8W
			jmp		short 3f
_adj_4020:
			__adj_fnt10W
			####	get font color
3:			movq	rdi,xmm0					# retrieve color palette
			and		ah,0x07						# pac.BRG
			mov		cl,ah
			movzx	rcx,cl
			shl		rcx,2
			mov		r13d,DWORD PTR[rdi+rcx]		# font color #1,#2
			mov		r14d,DWORD PTR[rdi+32+rcx]
			####	draw font/graph mix
			mov		rcx,r9						# retrieve lpitch
			movq	rdi,xmm6					# retrieve vram
			movq	xmm5,rdi					# save vram'
			movq	rax,xmm14					# retrieve address of _ras_802x
			jmp		rax
_ras_4020:
			__fntrastW
			__fntrastW
_ras_4025:
			.rept	8
			__fntrastW
			.endr

			movq	rbx,xmm6					# retrieve vram
			add		rbx,64						# next video addr
			movq	xmm6,rbx					# save vram
			movq	rsi,xmm1					# retrieve *pac
			movq	rdi,xmm2					# retrieve *txt
			pop		rcx							# retrieve col counter
			dec		cx
			jnz		20b

			####	end of draw
			pop		r15
			pop		r14
			pop		r13
			pop		rbx
			pop		rbp
			ret
