/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2021 H.Kanemoto
 **
 **		configurator
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <termio.h>
#include <stdint.h>
#include <iconv.h>
#include <linux/input.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#define	export
/**--------------------------------------------------------------------------
 **	external variable/functions
 **------------------------------------------------------------------------*/
extern void		draw_icon(LPDWORD video, int pitch, int x, int y);
extern void     draw_text(int line, LPBYTE attr, LPBYTE text);
extern void	    _n80_mbwrite(USHORT Adr, BYTE v, DWORD *W);
extern N80LX	n80lx;
/**--------------------------------------------------------------------------
 **	declare structures
 **------------------------------------------------------------------------*/
typedef void    *LPVOID;
/**--------------------------------------------------------------------------
 **	declare macro
 **------------------------------------------------------------------------*/
#define TYPE_TOGGLE     (0)     // toggle control
#define TYPE_TEXTIN     (1)     // text-in control
#define TYPE_DRVSELECT  (9)     // drive selector
#define TYPE_IMGSLOT  (100)     // disk image slot
#define TYPE_TERMINATOR (-1)    // terminator dummy control
#define VERB_INIT       (0)     // INIT control
#define VERB_EVENT      (1)     // Key Event handle
#define VERB_PRINT      (2)     // Print control
#define VERB_PRINT_SEL  (3)     // Print control (selected status)
#define VERB_COMMIT		(4)     // on toggled set value
#define VERB_TEXTIN_CH  (5)     // on key char
#define ACT_NONE        (0)
#define ACT_DONE        (1)
#define ACT_CHANGEPAGE  (9)
#define ACT_QUIT        (-1)
#define CHAR_VALID      (1)
#define CHAR_INVALID    (0)
#define SP              (32)
/**--------------------------------------------------------------------------
 **	local function declarationLPCFGITEM
 **------------------------------------------------------------------------*/
static int _controller_tenkeyless(LPCFGITEM, int, LPVOID);
static int _controller_interrace(LPCFGITEM, int, LPVOID);
//static int _controller_nowait(LPCFGITEM, int, LPVOID);
static int _controller_cmtfile(LPCFGITEM, int, LPVOID);
static int _controller_n80rom(LPCFGITEM, int, LPVOID);
static int _controller_fdurom(LPCFGITEM, int, LPVOID);
static int _controller_bdf(LPCFGITEM, int, LPVOID);
static int _controller_snddev(LPCFGITEM, int, LPVOID);
static int _controller_image_slot(LPCFGITEM, int, LPVOID);
static int _controller_drive_select(LPCFGITEM, int, LPVOID);
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static CFGITEM page0_items[] = {
	{0x01, TYPE_TOGGLE, "TEN-Keyless .......................... [ ]", "assign arrow key to Ten-Key (2/4/6/8) for Tenkey-less keyboard", 0, _controller_tenkeyless},
	{0x02, TYPE_TOGGLE, "Interlace-like effect ................ [ ]", "display like interlace", 0, _controller_interrace},
	//{0x03, TYPE_TOGGLE, "No-wait .............................. [ ]", "ignore memory wait, M1 wait", 0, _controller_nowait},
	{0x11, TYPE_TEXTIN, "N80 rom file name .................... [                        ]", "N-BASIC ROM file name (required)", 24, _controller_n80rom},
	{0x12, TYPE_TEXTIN, "Disk Unit rom file name .............. [                        ]", "Disk ROM file name (optional)", 24, _controller_fdurom},
	{0x31, TYPE_TEXTIN, "output cmt file name ................. [                 ]", "cmt output file name when csave or mon W command (required)", 17-1, _controller_cmtfile},
	{0x51, TYPE_TEXTIN, "kanji font file name ................. [                                ]", "Kanji 16x16 bitmap font (bdf) file name (required)", 32, _controller_bdf},
	{0x52, TYPE_TEXTIN, "alsa sound device name ............... [                                    ]", "last selected alsa sound device name", 36, _controller_snddev},
	{0x41, TYPE_IMGSLOT, "Image File Slot #0 Control ........... [ ]", "[DEL] remove [INS] load [RET] insert/eject disk image", 0, _controller_image_slot},
	{0x42, TYPE_IMGSLOT, "Image File Slot #1 Control ........... [ ]", "[DEL] remove [INS] load [RET] insert/eject disk image", 0, _controller_image_slot},
	{0x00, TYPE_TERMINATOR}
};
static CFGITEM page1_items[] = {
	{0x01, TYPE_DRVSELECT, " 0) .................................. [ ]", "inser/eject disk image #0 to/from drive", -1, _controller_drive_select},
	{0x02, TYPE_DRVSELECT, " 1) .................................. [ ]", "inser/eject disk image #1 to/from drive", -1, _controller_drive_select},
	{0x03, TYPE_DRVSELECT, " 2) .................................. [ ]", "inser/eject disk image #2 to/from drive", -1, _controller_drive_select},
	{0x04, TYPE_DRVSELECT, " 3) .................................. [ ]", "inser/eject disk image #3 to/from drive", -1, _controller_drive_select},
	{0x05, TYPE_DRVSELECT, " 4) .................................. [ ]", "inser/eject disk image #4 to/from drive", -1, _controller_drive_select},
	{0x06, TYPE_DRVSELECT, " 5) .................................. [ ]", "inser/eject disk image #5 to/from drive", -1, _controller_drive_select},
	{0x07, TYPE_DRVSELECT, " 6) .................................. [ ]", "inser/eject disk image #6 to/from drive", -1, _controller_drive_select},
	{0x08, TYPE_DRVSELECT, " 7) .................................. [ ]", "inser/eject disk image #7 to/from drive", -1, _controller_drive_select},
	{0x09, TYPE_DRVSELECT, " 8) .................................. [ ]", "inser/eject disk image #8 to/from drive", -1, _controller_drive_select},
	{0x0A, TYPE_DRVSELECT, " 9) .................................. [ ]", "inser/eject disk image #9 to/from drive", -1, _controller_drive_select},
};
static LPCFGITEM page_items[] = { page0_items, page1_items };
/**--------------------------------------------------------------------------
 **	local functions
 **------------------------------------------------------------------------*/
static char to_ascii(LPKEYEVT kv, SDL_Event *e) {
	switch (e->key.keysym.sym) {
	case SDLK_ESCAPE:	/*ESC*/	kv->key = 0x01; return 0x1B;
	case SDLK_BACKSPACE:/* BS*/	kv->key = 0x0E; return 0x08;
	case SDLK_TAB:		/*TAB*/ kv->key = 0x0F; return 0x09;
	case SDLK_RETURN:	/*RET*/	kv->key = 0x1C; return 0x0D;
	case SDLK_SPACE:	/* SP*/	kv->key = 0x39; return 0x20;
	case SDLK_1:		/* 1 */	kv->key = 0x02;	return kv->SHIFT ? 0x21 : 0x31;
	case SDLK_2:		/* 2 */	kv->key = 0x02;	return kv->SHIFT ? 0x22 : 0x32;
	case SDLK_3:		/* 3 */	kv->key = 0x02;	return kv->SHIFT ? 0x23 : 0x33;
	case SDLK_4:		/* 4 */	kv->key = 0x02;	return kv->SHIFT ? 0x24 : 0x34;
	case SDLK_5:		/* 5 */	kv->key = 0x02;	return kv->SHIFT ? 0x25 : 0x35;
	case SDLK_6:		/* 6 */	kv->key = 0x02;	return kv->SHIFT ? 0x26 : 0x36;
	case SDLK_7:		/* 7 */	kv->key = 0x02;	return kv->SHIFT ? 0x27 : 0x37;
	case SDLK_8:		/* 8 */	kv->key = 0x02;	return kv->SHIFT ? 0x28 : 0x38;
	case SDLK_9:		/* 9 */	kv->key = 0x02;	return kv->SHIFT ? 0x29 : 0x39;
	case SDLK_0:		/* 0 */	kv->key = 0x02;	return kv->SHIFT ? 0x30 : 0x30;

	case SDLK_q:		/* Q */	kv->key = 0x10; return kv->SHIFT ? 0x51 : 0x71;
	case SDLK_w:		/* W */	kv->key = 0x11;	return kv->SHIFT ? 0x57 : 0x77;
	case SDLK_e:		/* E */	kv->key = 0x12;	return kv->SHIFT ? 0x45 : 0x65;
	case SDLK_r:		/* R */	kv->key = 0x13;	return kv->SHIFT ? 0x52 : 0x72;
	case SDLK_t:		/* T */	kv->key = 0x14;	return kv->SHIFT ? 0x54 : 0x74;
	case SDLK_y:		/* Y */	kv->key = 0x15;	return kv->SHIFT ? 0x59 : 0x79;
	case SDLK_u:		/* U */	kv->key = 0x16;	return kv->SHIFT ? 0x55 : 0x75;
	case SDLK_i:		/* I */	kv->key = 0x17;	return kv->SHIFT ? 0x49 : 0x69;
	case SDLK_o:		/* O */	kv->key = 0x18;	return kv->SHIFT ? 0x4F : 0x6F;
	case SDLK_p:		/* P */	kv->key = 0x19;	return kv->SHIFT ? 0x50 : 0x70;

	case SDLK_a:		/* A */	kv->key = 0x1E;	return kv->SHIFT ? 0x41 : 0x61;
	case SDLK_s:		/* S */	kv->key = 0x1F;	return kv->SHIFT ? 0x53 : 0x73;
	case SDLK_d:		/* D */	kv->key = 0x20;	return kv->SHIFT ? 0x44 : 0x64;
	case SDLK_f:		/* F */	kv->key = 0x21;	return kv->SHIFT ? 0x46 : 0x66;
	case SDLK_g:		/* G */	kv->key = 0x22;	return kv->SHIFT ? 0x47 : 0x67;
	case SDLK_h:		/* H */	kv->key = 0x23;	return kv->SHIFT ? 0x48 : 0x68;
	case SDLK_j:		/* J */	kv->key = 0x24;	return kv->SHIFT ? 0x4A : 0x6A;
	case SDLK_k:		/* K */	kv->key = 0x25;	return kv->SHIFT ? 0x4B : 0x6B;
	case SDLK_l:		/* L */	kv->key = 0x26;	return kv->SHIFT ? 0x4C : 0x6C;

	case SDLK_z:		/* Z */	kv->key = 0x2C;	return kv->SHIFT ? 0x5A : 0x7A;
	case SDLK_x:		/* X */	kv->key = 0x2D;	return kv->SHIFT ? 0x58 : 0x78;
	case SDLK_c:		/* C */	kv->key = 0x2E;	return kv->SHIFT ? 0x43 : 0x63;
	case SDLK_v:		/* V */	kv->key = 0x2F;	return kv->SHIFT ? 0x56 : 0x76;
	case SDLK_b:		/* B */	kv->key = 0x30;	return kv->SHIFT ? 0x42 : 0x62;
	case SDLK_n:		/* N */	kv->key = 0x31;	return kv->SHIFT ? 0x4E : 0x6E;
	case SDLK_m:		/* M */	kv->key = 0x32;	return kv->SHIFT ? 0x4D : 0x6D;

	case SDLK_MINUS:		/* = -   */	kv->key = 0x0C;	return kv->SHIFT ? 0x3D : 0x2D;
	case SDLK_CARET:		/* ^ ~   */	kv->key = 0x0D;	return kv->SHIFT ? 0x7E : 0x5E;
	case SDLK_AT:			/* @     */	kv->key = 0x1A;	return kv->SHIFT ? 0x60 : 0x40;
	case SDLK_LEFTBRACKET:	/* [     */	kv->key = 0x1B;	return kv->SHIFT ? 0x7B : 0x5B;
	case SDLK_SEMICOLON:	/* + ;   */	kv->key = 0x27;	return kv->SHIFT ? 0x2B : 0x3B;
	case SDLK_COLON:		/* * :   */	kv->key = 0x28;	return kv->SHIFT ? 0x2A : 0x3A;
	case SDLK_RIGHTBRACKET:	/* ]     */	kv->key = 0x2B;	return kv->SHIFT ? 0x7D : 0x5D;
	case SDLK_COMMA:		/* < ,   */	kv->key = 0x33;	return kv->SHIFT ? 0x3C : 0x2C;
	case SDLK_PERIOD:		/* > .   */	kv->key = 0x34;	return kv->SHIFT ? 0x3E : 0x2E;
	case SDLK_SLASH:		/* / ?   */	kv->key = 0x35;	return kv->SHIFT ? 0x3F : 0x2F;
	case SDLK_BACKSLASH:	/* \ |   */	kv->key = 0x7C;	return kv->SHIFT ? 0x5C : 0x7C;
	case SDLK_UNDERSCORE:	/* _     */	kv->key = 0x59;	return kv->SHIFT ? 0x5F : 0x5C;

	case SDLK_INSERT:		/* INS   */	kv->key = VK_INSERT;	return '\0';
	case SDLK_DELETE:		/* DEL   */	kv->key = VK_DELETE;	return '\0';
	case SDLK_UP:			/* ARW-U */	kv->key = VK_UP;		return '\0';
	case SDLK_LEFT:			/* ARW-L */	kv->key = VK_LEFT;		return '\0';
	case SDLK_RIGHT:		/* ARW-R */	kv->key = VK_RIGHT;		return '\0';
	case SDLK_DOWN:			/* ARR-D */	kv->key = VK_DOWN;		return '\0';

	default:    						kv->key = 0x00;			return '\0';
	}
}
static bool is_alpha(LPKEYEVT kv) {
	if (kv->ascii >= 'a' && kv->ascii <= 'z')
		return true;
	if (kv->ascii >= 'A' && kv->ascii <= 'Z')
		return true;
	return false;
}
static bool is_numeric(LPKEYEVT kv) {
	return (kv->ascii >= '0' && kv->ascii <= '9');
}
static void _toggle_print(LPCFGITEM self, char *vram, char on) {
	memset(vram, SP, 78);
	strcpy(vram, self->name);
	if (self->value == 1)
		vram[40] = on;
}
static int _controller_common_toggle(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
				self->value = self->value == 1 ? 0 : 1;
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_DONE;
			} else
				return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL:
			_toggle_print(self, (char*)param, 'X');
			break;
	}
	return ACT_NONE;
}
static int _controller_tenkeyless(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80lx.tenkeyless;
			break;
		case VERB_COMMIT:
			n80lx.tenkeyless = self->value;
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_interrace(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80device.param.interlace ? 1 : 0;
			break;
		case VERB_COMMIT:
			n80device.param.interlace = self->value == 1;
			n80device.blt.chg_palette(n80device.param.interlace);
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
/*static int _controller_nowait(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80lx.cpu_no_wait ? 1 : 0;
			break;
		case VERB_COMMIT:
			n80lx.cpu_no_wait = self->value == 1;
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}*/
static int _controller_common_filename(LPCFGITEM self, int verb, LPVOID param, char *buff, const char *defval) {
	switch (verb) {
		case VERB_INIT:
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_RETURN) {
				self->controller(self, VERB_COMMIT, kv);
				return ACT_DONE;
			}
			int len = strlen(buff);
			if (len > self->value)
				return ACT_NONE;
			if (kv->key == VK_BACK && len > 0) {    // [BS] Key pressed
				buff[len - 1] = '\0';
				return ACT_DONE;
			} else if (len < self->value && self->controller(self, VERB_TEXTIN_CH, kv) == CHAR_VALID) {
				buff[len] = kv->ascii;
				buff[len+1] = '\0';
				return ACT_DONE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL: {
			//_textin_print(self, (char*)param, buff, self->value, verb == VERB_PRINT_SEL);
			//break;
			char *vram = (char*)param;
			int l = strlen(buff);
			memset(vram, SP, 78);
			strcpy(vram, self->name);
			for (int i = 0; i <  self->value; ++i) {
				char ch = *buff++;
				if (ch == '\0')
					break;
				vram[40 + i] = ch;
			}
			// show guide cursor when selected
			if (verb == VERB_PRINT_SEL) {
				if (l <  self->value)
					vram[40 + l] = 0x1D;
			}
		}
		case VERB_TEXTIN_CH: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (is_alpha(kv))
				return CHAR_VALID;
			if (is_numeric(kv))
				return CHAR_VALID;
			if (kv->ascii == '_' || kv->ascii == '.')
				return CHAR_VALID;
			if (kv->ascii == '-' && strlen(buff) > 0)
				return CHAR_VALID;
			return CHAR_INVALID;
		}
		case VERB_COMMIT:
			if (strlen(buff) == 0 && defval != NULL)
				strcpy(buff, defval);
			break;
	}
	return ACT_NONE;
}
static int _controller_cmtfile(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.save_cmt_name, "casette.cmt");
}
static int _controller_n80rom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.n80rom_filename, "PC-8001.ROM");
}
static int _controller_fdurom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.fdurom_filename, NULL);
}
static int _controller_bdf(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.bdf_filename, "shnmk16.bdf");
}
static int _controller_snddev(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.snd_dev_name, NULL);
}
static int _controller_image_slot(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = self->arg == NULL ? 0 : 1;
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (self->value == 1) {
				if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
					self->parent->jmp_arg  = self->arg;
					self->parent->jmp_page = 1;
					return ACT_CHANGEPAGE;
				} else if (kv->key == VK_DELETE) {
					int slot_no = (self->id & 0x0f) - 1;
					n80device.fdu.mgr.detach(&n80device.fdu.mgr, slot_no);
					self->arg   = NULL;
					self->value = 0;
					return ACT_DONE;
				}
			} else if (kv->key == VK_INSERT) {
				int slot_no = (self->id & 0x0f) - 1;
				n80lx.chooser.open(&n80lx.chooser, slot_no, -1);
				self->parent->jmp_arg  = NULL;
				self->parent->jmp_page = 0;
				return ACT_CHANGEPAGE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL:
			_toggle_print(self, (char*)param, '\x1c');
			break;
		case VERB_COMMIT:
			break;
	}
	return ACT_NONE;
}
static int _controller_drive_select(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = ((LPFLOPPYDISK)self->arg)->unit_no;
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_CHANGEPAGE;
			} else if (kv->key == VK_LEFT) {
				switch (self->value) {
					case 0:	self->value = -1;	break;
					case 1: self->value = 0;	break;
					default:self->value = 1;	break;
				}
				return ACT_DONE;
			} else if (kv->key == VK_RIGHT) {
				switch (self->value) {
					case 0:	self->value = 1;	break;
					case 1: self->value = -1;	break;
					default:self->value = 0;	break;
				}
				return ACT_DONE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL: {
			LPFLOPPYDISK disk = (LPFLOPPYDISK) self->arg;
			char *vram = (char*)param;
			memset(vram, SP, 78);
			strcpy(vram, self->name);
			// copy disk name
			char *title = disk->header->title[0] == 0 ? "{NO TITLE}\0\0\0\0\0\0\0" : disk->header->title;
			for (int j = 0; j < 17; ++j) {
				char c = title[j];
				vram[j + 4] = c == '\0' ? '.' : c;
			}
			// mount status
			char c = ' ';
			switch (self->value) {
				case 0: c = 'A'; break;
				case 1: c = 'B'; break;
			}
			vram[40] = c;
			break;
		}
		case VERB_COMMIT: {
			LPFLOPPYDISK disk = (LPFLOPPYDISK) self->arg;
			if (disk->unit_no != self->value) {
				if (disk->unit_no == 0 || disk->unit_no == 1)
					n80device.fdu.fdd.eject(&n80device.fdu.fdd, disk->unit_no);
				if (self->value == 0 || self->value ==  1) {
					n80device.fdu.fdd.eject(&n80device.fdu.fdd, self->value);
					n80device.fdu.fdd.insert(&n80device.fdu.fdd, self->value, disk);
				}
				self->value = disk->unit_no;
				self->parent->reflesh(self->parent);
			}
			break;
		}
	}
	return ACT_NONE;
}
/**--------------------------------------------------------------------------
 **	page methods
 **------------------------------------------------------------------------*/
static void drawitem(LPCFGITEM item, int line, BOOL selected) {
	int row = line + 9;
	item->controller(item, selected ? VERB_PRINT_SEL : VERB_PRINT, &n80lx.text[row][1]);
	for (int i = 1; i < 79; ++i)
		n80lx.attr[row][i] = selected ? 0x22 : 0x07;
	if (selected) {
		memset(&n80lx.text[7][1], SP, 78);
		sprintf(&n80lx.text[7][1], " hint: %s", item->hint);
	}
}

static LPCFGITEM get_item(LPCFGPAGE self) {
	return &self->items[self->cur_idx];
}

static int call_current_item(LPCFGPAGE self, int verb, LPVOID arg) {
	LPCFGITEM item = self->get_item(self);
	return item->controller(item, verb, arg);
}

static void page_reflesh(LPCFGPAGE self) {
	for (int i = 0; i < self->item_cnt; ++i) {
		LPCFGITEM item = &self->items[i];
		item->controller(item, VERB_INIT, NULL);
		self->parent->drawitem(item, i, i == self->cur_idx);
	}
	self->parent->show(self->parent);
}

static void prepare_page(LPCFGPAGE self, LPVOID arg) {
	self->item_cnt = 0;
	self->cur_idx = 0;
	switch (self->page) {
		case 0:
			for (int r = 0; self->items[r].type != TYPE_TERMINATOR; ++r) {
				self->items[r].parent = self;
				self->item_cnt++;
				if (self->items[r].type == TYPE_IMGSLOT) {
					int slot = (self->items[r].id & 0x0f) - 1;
					self->items[r].arg = n80device.fdu.mgr.slots[slot];
				}
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(&self->items[r], r, r == 0);
			}
			break;
		default: {
			LPDISKIMAGE image	= (LPDISKIMAGE)arg;
			for (int r = 0; r < image->disk_count; ++r) {
				LPFLOPPYDISK disk = image->disks[r];
				LPCFGITEM item = &self->items[r];
				item->parent = self;
				item->arg = disk;
				self->item_cnt++;
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(item, r, r == 0);
			}
		}
	}
}

static int handle_evt_page(LPCFGPAGE self, LPKEYEVT kv) {
	LPCFGITEM item = self->get_item(self);
	int action = item->controller(item, VERB_EVENT, kv);
	switch (self->page) {
		case 0:
			if (kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				n80lx.ini.flush(&n80lx.ini);
				return ACT_QUIT;
			}
			break;
		default: {
			if (action == ACT_QUIT || kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				self->jmp_arg  = NULL;
				self->jmp_page = 0;
				return ACT_CHANGEPAGE;
			}
			break;
		}
	}
	return action;
}
/**--------------------------------------------------------------------------
 **	config methods
 **------------------------------------------------------------------------*/
static void show(LPCONFIG self) {
	int r;
	for (r = 0; r < 25; ++r)
		draw_text(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
	draw_icon(n80lx.video, 640, 12, 16);
	SDL_UpdateTexture(n80lx.texture, NULL, n80lx.video, 640 * 4);
	SDL_RenderClear(n80lx.renderer);
	SDL_RenderCopy(n80lx.renderer, n80lx.texture, NULL, NULL);
	SDL_RenderPresent(n80lx.renderer);
}

static void init_page(LPCONFIG self, int page, LPVOID arg) {
	self->page		= page;
	self->maxpages	= 2;

	// init page objects
	for (int p = 0; p < self->maxpages; ++p) {
		self->pages[p].parent	= self;
		self->pages[p].page		= p;
		self->pages[p].items	= page_items[p];
		self->pages[p].get_item	= get_item;
		self->pages[p].call		= call_current_item;
		self->pages[p].reflesh	= page_reflesh;
		self->pages[p].prepare  = prepare_page;
		self->pages[p].handler  = handle_evt_page;
	}
	// clear and draw base screen 
	for (int r = 0; r < 25; ++r) {
		for (int c = 0; c < 80; ++c) {
			BYTE atr = 0x23;
			if ((r == 4 || r == 5) && (c > 0 && c < 79))
				atr = 0x27;
			else if (r == 7 && (c > 0 && c < 79))
				atr = 0x26;
			else if ((r > 8 && r < 24) && (c > 0 && c < 79))
				atr = 0x07;
			n80lx.attr[r][c] = atr;
			n80lx.text[r][c] = SP;
		}
	}
	const char *title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for Linux (SDL2)";
	const char *title2 = "    Copyright 1993-2021 (C) H.Kanemoto, All Right Reserved.";
	const char *title3 = "System Config";
	const char* title4 = "usage:  [ESC] Quit/Back, [\x1e\x1f] Move, [SP] Toggle, [\x1d\x1c] Value";
	const char *title5 = "hint: ";
	memcpy(&n80lx.text[1][2], title1, strlen(title1));
	memcpy(&n80lx.text[2][2], title2, strlen(title2));
	memcpy(&n80lx.text[4][2], title3, strlen(title3));
	memcpy(&n80lx.text[5][2], title4, strlen(title4));
	memcpy(&n80lx.text[7][2], title5, strlen(title5));

	self->pages[page].prepare(&self->pages[page], arg);
}

static void evt_pump(LPCONFIG self) {
	SDL_Event e;
	memset(&self->kv, 0, sizeof(KEYEVTG));
	for (bool quit = false; !quit;) {
		while (SDL_PollEvent(&e)) {
			if (e.type == SDL_QUIT)
				quit = true;
			else if (e.type == SDL_KEYUP) {
				switch (e.key.keysym.sym) {
					case SDLK_LSHIFT:
					case SDLK_RSHIFT:	self->kv.SHIFT = false; break;
					case SDLK_LALT:
					case SDLK_RALT:		self->kv.ALT   = false; break;
				}
			} else if (e.type == SDL_KEYDOWN) {
				switch (e.key.keysym.sym) {
					case SDLK_LSHIFT:
					case SDLK_RSHIFT:	self->kv.SHIFT = true; break;
					case SDLK_LALT:
					case SDLK_RALT:		self->kv.ALT   = true; break;
					default: {
						self->kv.ascii = to_ascii(&self->kv, &e);
						quit = self->dispatch(self, &self->kv);
						break;
					}
				}
			}
		}
	}
}

static BOOL dispatch(LPCONFIG self, LPKEYEVT kv) {
	LPCFGPAGE page = &self->pages[self->page];

	int action = page->handler(page, kv);
	if (action == ACT_DONE) {
		self->drawitem(page->get_item(page), page->cur_idx, true);
		self->show(self);
	} else if (action == ACT_CHANGEPAGE) {
		self->init_page(self, page->jmp_page, page->jmp_arg);
		self->show(self);
	} else if (action == ACT_QUIT) {
		return true;
	}
	else if ((kv->key == VK_TAB && !kv->SHIFT) || kv->key == VK_DOWN) { // [TAB] move to next item
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (++page->cur_idx >= page->item_cnt)
				page->cur_idx = 0;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	} else if ((kv->key == VK_TAB && kv->SHIFT) || kv->key == VK_UP) {
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (--page->cur_idx < 0)
				page->cur_idx = page->item_cnt - 1;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	}
	return false;
}

static void config(LPCONFIG self) {
	self->init_page(self, 0, NULL);
	self->show(self);
	self->evt_pump(self);
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void init_config(LPCONFIG self) {
	self->init		= init_config;
	self->open      = config;
	self->init_page = init_page;
	self->evt_pump	= evt_pump;
	self->dispatch	= dispatch;
	self->to_ascii  = to_ascii;
	self->drawitem	= drawitem;
	self->show		= show;
}
