/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		frame buffer operations
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#define	export
/**--------------------------------------------------------------------------
 **	外部変数／関数
 **------------------------------------------------------------------------*/
extern N80LX		n80lx;
export const USHORT	SPITZ_COLOR[8] = {0x0000,0x001f,0x07e0,0x07ff,0xf800,0xf81f,0xffe0,0xffff};
/**--------------------------------------------------------------------------
 **	内部変数／関数
 **------------------------------------------------------------------------*/
static void 		blt4020(int line);
static void 		blt4025(int line);
static void 		blt8020(int line);
static void 		blt8025(int line);
typedef void 		(*BLTFNC)(int);
static BLTFNC		bltfncs[] = {blt4020,blt4025,blt8020,blt8025};
static BLTFNC		fnc = blt4020;
/**--------------------------------------------------------------------------
 **	N80 main thread
 **------------------------------------------------------------------------*/
#ifdef _ZAURUS_
static void rotate_font_clockwise90(LPBYTE font)
{
	BYTE	bit = 0x80, tmp[8];
	int		i;

	for (i = 0; i < 8; ++i) {
		BYTE rot = 0;
		rot |= *(font+0) & bit ? 0x01 : 0;
		rot |= *(font+1) & bit ? 0x02 : 0;
		rot |= *(font+2) & bit ? 0x04 : 0;
		rot |= *(font+3) & bit ? 0x08 : 0;
		rot |= *(font+4) & bit ? 0x10 : 0;
		rot |= *(font+5) & bit ? 0x20 : 0;
		rot |= *(font+6) & bit ? 0x40 : 0;
		rot |= *(font+7) & bit ? 0x80 : 0;

		tmp[i] = rot;
		bit >>= 1;
	}
	for (i = 0; i < 8; ++i)
		*font++ = tmp[i];
}
#endif
void clr_fb_display(LPVOID vram)
{
	__asm__ __volatile__ (
	"		ldr		r0,%0			\n\t"
	"		mov		r1,#6400		\n\t"
	"		wxor	wr0,wr0,wr0		\n\t"
	"		wmov	wr1,wr0			\n\t"
	"		wmov	wr2,wr0			\n\t"
	"		wmov	wr3,wr0			\n\t"
	"		wmov	wr4,wr0			\n\t"
	"		wmov	wr5,wr0			\n\t"
	"		wmov	wr6,wr0			\n\t"
	"		wmov	wr7,wr0			\n\t"
	"		wmov	wr8,wr0			\n\t"
	"		wmov	wr9,wr0			\n\t"
	"		wmov	wr10,wr0		\n\t"
	"		wmov	wr11,wr0		\n\t"
	"0:		wstrd	wr0,[r0],#8		\n\t"
	"		wstrd	wr1,[r0],#8		\n\t"
	"		wstrd	wr2,[r0],#8		\n\t"
	"		wstrd	wr3,[r0],#8		\n\t"
	"		wstrd	wr4,[r0],#8		\n\t"
	"		wstrd	wr5,[r0],#8		\n\t"
	"		wstrd	wr6,[r0],#8		\n\t"
	"		wstrd	wr7,[r0],#8		\n\t"
	"		wstrd	wr8,[r0],#8		\n\t"
	"		wstrd	wr9,[r0],#8		\n\t"
	"		wstrd	wr10,[r0],#8	\n\t"
	"		wstrd	wr11,[r0],#8	\n\t"
	"		subs	r1,r1,#1		\n\t"
	"		bne		0b				\n\t"
	:
	: "m"(vram)
	: "r0","r1","wr0","wr1","wr2","wr3","wr4","wr5","wr6","wr7","wr8","wr9","wr10","wr11"
	);
}

static inline void blt80(LPVOID src, LPVOID dst)
{
	__asm__ __volatile__ (
	"		ldr		r0,%0			\n\t"
	"		ldr		r1,%1			\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wldrd	wr0,[r0]		\n\t"
	"		wstrd	wr0,[r1]		\n\t"
	:
	: "m"(src), "m"(dst)
	: "r0","r1","wr0"
	);
}
static inline void blt40(LPVOID src, LPVOID dst)
{
	__asm__ __volatile__ (
	"		ldr		r0,%0			\n\t"
	"		ldr		r1,%1			\n\t"
	"		add		r2,r1,#960		\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wstrd	wr0,[r2],+#8	\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wstrd	wr0,[r2],+#8	\n\t"
	"		wldrd	wr0,[r0],+#8	\n\t"
	"		wstrd	wr0,[r1],+#8	\n\t"
	"		wstrd	wr0,[r2],+#8	\n\t"
	"		wldrd	wr0,[r0]		\n\t"
	"		wstrd	wr0,[r1]		\n\t"
	"		wstrd	wr0,[r2]		\n\t"
	:
	: "m"(src), "m"(dst)
	: "r0","r1","r2","wr0"
	);
}

export int init_font()
{
	FILE *fp;
	int	i, len;
	
	// font file load
	fp = fopen("pc-8001.fon", "r");
	if (fp == NULL) {
		fprintf(stderr, "can not open font file\n");
		return 1;
	}
	len = fread(&n80device.font, 1, (1024*4), fp);
	fclose(fp);
	if (len != (1024*4)) {
		fprintf(stderr, "invalid font file size\n");
		return 1;
	}
	#ifdef _ZAURUS_
	// フォントデータを右90度回転
	for (i = 0; i < 512; ++i)
		rotate_font_clockwise90(&n80device.font[i*8]);
	#endif
	// フォントデータをPCGにコピー
	for (i = 0; i < 2048; ++i)
		n80device.pcg[i] = n80device.font[i];

	// fb用ビットパターンデータを作成
	for (i = 0; i < 8; ++i) {
		USHORT color = SPITZ_COLOR[i];
		int j;
		for (j = 0; j < 256; ++j) {
			BYTE bit = (BYTE) (j & 0xff);
			n80lx.dot_pattern[i][j][0]  = bit & 0x80 ? color : 0;
			n80lx.dot_pattern[i][j][1]  = bit & 0x80 ? color : 0;
			n80lx.dot_pattern[i][j][2]  = bit & 0x40 ? color : 0;
			n80lx.dot_pattern[i][j][3]  = bit & 0x40 ? color : 0;
			n80lx.dot_pattern[i][j][4]  = bit & 0x20 ? color : 0;
			n80lx.dot_pattern[i][j][5]  = bit & 0x20 ? color : 0;
			n80lx.dot_pattern[i][j][6]  = bit & 0x10 ? color : 0;
			n80lx.dot_pattern[i][j][7]  = bit & 0x10 ? color : 0;
			n80lx.dot_pattern[i][j][8]  = bit & 0x08 ? color : 0;
			n80lx.dot_pattern[i][j][9]  = bit & 0x08 ? color : 0;
			n80lx.dot_pattern[i][j][10] = bit & 0x04 ? color : 0;
			n80lx.dot_pattern[i][j][11] = bit & 0x04 ? color : 0;
			n80lx.dot_pattern[i][j][12] = bit & 0x02 ? color : 0;
			n80lx.dot_pattern[i][j][13] = bit & 0x02 ? color : 0;
			n80lx.dot_pattern[i][j][14] = bit & 0x01 ? color : 0;
			n80lx.dot_pattern[i][j][15] = bit & 0x01 ? color : 0;
		}
	}
	return 0;
}

export void enter()
{
	pthread_mutex_lock(&n80lx.vram_sync);
}

export void leave()
{
	pthread_mutex_unlock(&n80lx.vram_sync);
}

export void	chg_bltfunc(BYTE mode)
{
	fnc = bltfncs[mode];
}

static void blt4020(int line)
{
	if (line < 20) {
		int			col, r;
		LPUSHORT	vid  = n80lx.video + 40 + (19 - line) * 20;
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];

		for (col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			USHORT		udr1, udr2, upr1, upr2, rgb = SPITZ_COLOR[color];

			txts += 2,	pacs += 2;
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			//  Normal      Reverse
			//  Under None  Under None
			//   *o    oo    o*    **
			if (attr & 0x20) {
				if (attr & 0x80) {
					udr1 = 0;
					udr2 = rgb;
				} else {
					udr1 = rgb;
					udr2 = rgb;
				}
			} else {
				if (attr & 0x80) {
					udr1 = rgb;
					udr2 = 0;
				} else {
					udr1 = 0;
					udr2 = 0;
				}
			}
			//  Normal      Reverse
			//  Upper None  Upper None
			//   *o    oo    o*    **
			if (attr & 0x20) {
				if (attr & 0x08) {
					upr1 = rgb;
					upr2 = 0;
				} else {
					upr1 = rgb;
					upr2 = rgb;
				}
			} else {
				if (attr & 0x08) {
					upr1 = 0;
					upr2 = rgb;
				} else {
					upr1 = 0;
					upr2 = 0;
				}
			}
			for (r = 0; r < 8; ++r) {
				BYTE	 bit = *fnt++;
				LPUSHORT ptr, v = vid;

				// lower space
				*(v+480) = udr1, *v++ = udr1;
				*(v+480) = udr2, *v++ = udr2;

				if (attr & 0x20)	// reverse
					bit = ~bit;
				ptr = &n80lx.dot_pattern[color][bit][0];
				blt40(ptr, v);
				v += 16;

				// upper space
				*(v+480) = upr1, *v++ = upr1;
				*(v+480) = upr2, *v   = upr2;

				vid += 960;
			}
		}
	}
}

static void blt4025(int line)
{
	if (line < 25) {
		int			col, r;
		LPUSHORT	vid  = n80lx.video + 40 + (24 - line) * 16;
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];

		for (col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;

			txts += 2, pacs += 2;
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			for (r = 0; r < 8; ++r) {
				BYTE	 bit = *fnt++;
				LPUSHORT ptr;

				if (attr & 0x80)	// under line
					bit |= 0x80;
				if (attr & 0x08)	// upper line
					bit |= 0x01;
				if (attr & 0x20)	// reverse
					bit = ~bit;

				ptr = &n80lx.dot_pattern[color][bit][0];
				blt40(ptr, vid);
				vid += 960;
			}
		}
	}
}

static void blt8020(int line)
{
	if (line < 20) {
		int			col, r;
		LPUSHORT	vid  = n80lx.video + 40 + (19 - line) * 20;
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];

		for (col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			USHORT		udr1, udr2, upr1, upr2, rgb = SPITZ_COLOR[color];

			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			//  Normal      Reverse
			//  Under None  Under None
			//   *o    oo    o*    **
			if (attr & 0x20) {
				if (attr & 0x80) {
					udr1 = 0;
					udr2 = rgb;
				} else {
					udr1 = rgb;
					udr2 = rgb;
				}
			} else {
				if (attr & 0x80) {
					udr1 = rgb;
					udr2 = 0;
				} else {
					udr1 = 0;
					udr2 = 0;
				}
			}
			//  Normal      Reverse
			//  Upper None  Upper None
			//   *o    oo    o*    **
			if (attr & 0x20) {
				if (attr & 0x08) {
					upr1 = rgb;
					upr2 = 0;
				} else {
					upr1 = rgb;
					upr2 = rgb;
				}
			} else {
				if (attr & 0x08) {
					upr1 = 0;
					upr2 = rgb;
				} else {
					upr1 = 0;
					upr2 = 0;
				}
			}
			for (r = 0; r < 8; ++r) {
				BYTE	 bit = *fnt++;
				LPUSHORT ptr, v = vid;

				// lower space
				*v++ = udr1;
				*v++ = udr2;

				if (attr & 0x20)	// reverse
					bit = ~bit;
				ptr = &n80lx.dot_pattern[color][bit][0];
				blt80(ptr, v);
				v += 16;

				// upper space
				*v++ = upr1;
				*v   = upr2;

				vid += 480;
			}
		}
	}
}

static void blt8025(int line)
{
	if (line < 25) {
		int			col, r;
		LPUSHORT	vid  = n80lx.video + 40 + (24 - line) * 16;
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];

		for (col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;

			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			for (r = 0; r < 8; ++r) {
				BYTE	 bit = *fnt++;
				LPUSHORT ptr;

				if (attr & 0x80)	// under line
					bit |= 0x80;
				if (attr & 0x08)	// upper line
					bit |= 0x01;
				if (attr & 0x20)	// reverse
					bit = ~bit;

				ptr = &n80lx.dot_pattern[color][bit][0];
				blt80(ptr, vid);
				vid += 480;
			}
		}
	}
}

export void reflesh_screen()
{
	int i, change_attr = 0, change_prev = 0, work_lines = 0;

	for (i = 0; i < 25; ++i) {
		n80device.vram_sync_enter();
		if (n80device.attr._attrchg[i] == 1 || change_prev) {
			n80device.attr.cache(i);
			change_attr = 1;
		} else
			change_attr = 0;
		n80device.vram_sync_leave();

		if (change_attr)
			change_prev = n80device.attr.eval(i) == 1 ? 1 : 0;
		if (change_attr) {
			fnc(i);
			++work_lines;
		}
	}
}
