/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		file chooser
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <termio.h>
#include <stdint.h>
#include <iconv.h>
#include <linux/input.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
/**--------------------------------------------------------------------------
 **	macro declare
 **------------------------------------------------------------------------*/
#define	export
#define ITEMARRAYUNIT	(15*3)
#define SP_VAL			(0x7f3e)
#define PC_VAL			(0xff3d)
#define STDSIZE			(32576)
#define MAXSIZE			(32768)
#define EMPTY_FILE  0
#define CURRENT_DIR 1
#define SOME_DIR	2
#define N80_FILE	3
#define CMT_FILE	4
#define T88_FILE	5
#define D88_FILE	6
/**--------------------------------------------------------------------------
 **	S T R U C T U R E
 **------------------------------------------------------------------------*/
typedef struct tagITEM {
	char*	name;
	int		type;
}	ITEM, *LPITEM;
/**--------------------------------------------------------------------------
 **	external variables/functions
 **------------------------------------------------------------------------*/
extern void		draw_icon(LPDWORD video, int pitch, int x, int y);
extern void		draw_text(int line, LPBYTE attr, LPBYTE text);
extern void		draw_textk(int line, LPBYTE attr, LPBYTE text);
extern N80LX	n80lx;
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static char		utf_buff[1024];
static char		sjis_buff[1024];
//static char		text[25][80];
//static char		attr[25][80];
static ITEM *	disp[15][3];
static BYTE		cmt_buff[1024*64];
/**--------------------------------------------------------------------------
 **	local functions
 **------------------------------------------------------------------------*/
static char *utf8to932(const char *utf) {
	memset(utf_buff, 0, 1024);
	memset(sjis_buff, 0, 1024);
	char *ibuf = utf_buff;
	char *obuf = sjis_buff;
	size_t inz = 1024, outz = 1024;

	iconv_t ic = iconv_open("SJIS", "UTF-8");
	if (ic == (iconv_t)-1) {
		fprintf(stderr, "can not iconv_open\n");
	} else {
		memcpy(utf_buff, utf, strlen(utf));
		iconv(ic, &ibuf, &inz, &obuf, &outz); 
		iconv_close(ic);
	}
	return sjis_buff;
}
static void show_file(int row, int col, LPITEM item, int selected) {
	int i = 0, c = 26 * col + 1;
	for (i = 0; i < 26; ++i) {
		n80lx.text[row + 9][c + i] = 0x20;
		char a = 0x02;
		switch (item != NULL ? item->type : EMPTY_FILE) {
			case CURRENT_DIR:
			case SOME_DIR:
				a = 0x05;
				break;
		}
		if (selected)
			a |= 0x20;
		n80lx.attr[row + 9][c + i] = a;
	}
	if (item == NULL)
		return;

	char *sjis = utf8to932(item->name);
	for (i = 0; i < 26 - 1; ++i) {
		n80lx.text[row + 9][c + i] = sjis[i];
	}
}
static void show(ITEM items[], int contents, int page, int index) {
	int r, c, n = 0;
	for (c = 0; c < 3; ++c) {
		for (r = 0; r < 15; ++r) {
			n = page * 45 + c * 15 + r;
			if (n < contents)
				disp[r][c] = &items[n];
			else
				disp[r][c] = NULL;
			show_file(r, c, disp[r][c], n == index ? 1 : 0);
		}
	}
	for (r = 0; r < 25; ++r) {
		if (r == 7 || (r > 8 && r < 24))
			draw_textk(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
		else
			draw_text(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
	}
	draw_icon(n80lx.video, 640, 12, 16);
	SDL_UpdateTexture(n80lx.texture, NULL, n80lx.video, 640 * 4);
	SDL_RenderClear(n80lx.renderer);
	SDL_RenderCopy(n80lx.renderer, n80lx.texture, NULL, NULL);
	SDL_RenderPresent(n80lx.renderer);
}
static int menu_select(LPITEM items, int contents) {
	int	index = 0, select = -1, upd_flag = 1;
	int	page = 0, r = 0, c = 0;

	SDL_Event e;
	for (int quit = 0; !quit;) {
		if (upd_flag == 1) {
			upd_flag = 0;
			index = page * 45 + c * 15 + r;
			show(items, contents, page, index);
		}
		while (SDL_PollEvent(&e)) {
			//if (quit)
			//	break;
			if (e.type == SDL_QUIT)
				quit = 1;
			else if (e.type == SDL_KEYDOWN) {
				switch (e.key.keysym.sym) {
				case SDLK_ESCAPE:
					quit = 1;
					break;
				case SDLK_DOWN:	// down-allow key
					if ((r + 1) < 15) {
						if (disp[r + 1][c] != NULL)
							++r;
					} else if ((c + 1) < 3) {
						if (disp[0][c + 1] != NULL) {
							r = 0;
							c++;
						}
					} else if (contents > ((page + 1) * 45)) {
						++page;
						r = 0;
						c = 0;
					}
					upd_flag = 1;
					break;
				case SDLK_UP:	// up-allow key
					if (r > 0) {
						--r;
					} else {
						if (c > 0) {
							--c;
							r = 14;
						} else if (page > 0) {
							--page;
							r = 14;
							c = 2;
						}
					}
					upd_flag = 1;
					break;
				case SDLK_PAGEDOWN:
					if (contents > ((page + 1) * 45)) {
						++page;
						r = 0;
						c = 0;
						upd_flag = 1;
					}
					break;
				case SDLK_PAGEUP:
					if (page > 0) {
						--page;
						r = 14;
						c = 2;
						upd_flag = 1;
					}
					break;
				case SDLK_RIGHT:	// right-allow key
					if ((c + 1) < 3) {
						if (disp[r][c + 1] != NULL)
							c++;
					} else
						c = 0;
					upd_flag = 1;
					break;
				case SDLK_LEFT:	// left-allow key
					if (c > 0)
						--c;
					else if (disp[r][2] != NULL)
						c = 2;
					upd_flag = 1;
					break;
				case SDLK_RETURN:
				case SDLK_SPACE:
					if (e.key.repeat == 0) {
						select = index;
						quit = 1;
					}
					break;
				}
			}
		}
	}
	return select;
}
static LPITEM enum_dir(const char *mydir, bool d88only, int *item_count) {
	DIR				*dir;
	struct dirent	*dp;
	ITEM			*items;
	int				slots = ITEMARRAYUNIT, contents = 0, remain = ITEMARRAYUNIT;

	*item_count = 0;
	items = (ITEM*) calloc(ITEMARRAYUNIT, sizeof(ITEM));
	if (items == NULL)
		return NULL;
	if ((dir = opendir(mydir)) != NULL) {
		for (dp = readdir(dir); dp != NULL; dp = readdir(dir)) {
			char *ptr;
			
			// 配列を拡張する
			if (remain == 0) {
				int extend = slots + ITEMARRAYUNIT;
				ITEM *new_items;
				new_items = (ITEM*) realloc(items, sizeof(ITEM)*extend);
				if (new_items == NULL)
					break;
				else
					items = new_items;
				slots += ITEMARRAYUNIT;
				remain = ITEMARRAYUNIT;
			}
			switch (dp->d_type) {
			case DT_DIR:
				items[contents].name = strdup(dp->d_name);
				if (!strcmp(dp->d_name, "."))
					items[contents].type = CURRENT_DIR;
				else
					items[contents].type = SOME_DIR;
				++contents;
				--remain;
				break;

			case DT_REG:
				if (d88only) {
					if ((ptr = strstr(dp->d_name, ".D88")) != NULL ||
						(ptr = strstr(dp->d_name, ".d88")) != NULL) {
						items[contents].name = strdup(dp->d_name);
						items[contents].type = D88_FILE;
						++contents;
						--remain;
					}
				} else {
					if ((ptr = strstr(dp->d_name, ".N80")) != NULL ||
						(ptr = strstr(dp->d_name, ".n80")) != NULL) {
						items[contents].name = strdup(dp->d_name);
						items[contents].type = N80_FILE;
						++contents;
						--remain;
					} else if ((ptr = strstr(dp->d_name, ".CMT")) != NULL ||
								(ptr = strstr(dp->d_name, ".cmt")) != NULL) {
						items[contents].name = strdup(dp->d_name);
						items[contents].type = CMT_FILE;
						++contents;
						--remain;
					} else if ((ptr = strstr(dp->d_name, ".T88")) != NULL ||
								(ptr = strstr(dp->d_name, ".t88")) != NULL) {
						items[contents].name = strdup(dp->d_name);
						items[contents].type = T88_FILE;
						++contents;
						--remain;
					} else if ((ptr = strstr(dp->d_name, ".D88")) != NULL ||
								(ptr = strstr(dp->d_name, ".d88")) != NULL) {
						items[contents].name = strdup(dp->d_name);
						items[contents].type = D88_FILE;
						++contents;
						--remain;
					}
				}
				break;
			}
		}
		closedir(dir);
	} else
		fprintf(stderr, "<%s> dir can not open\n", mydir);
		
	*item_count = contents;
	return items;
}
static int item_sort(const void *item1, const void *item2) {
	LPITEM i1 = (LPITEM)item1;
	LPITEM i2 = (LPITEM)item2;
	if (i1->type == i2->type)
		return strcmp(i1->name, i2->name);
	else if (i1->type > i2->type)
		return 1;
	else
		return -1;
}
static char *domenu(const char *mydir, bool d88only, int *isdir, int *type) {
	char	*result = NULL;
	int		contents = 0;

	LPITEM items = enum_dir(mydir, d88only, &contents);
	if (items == NULL)
		return result;

	if (contents > 0) {
		int i;
		char *sjis = utf8to932(mydir);
		for (i = 0; i < 72; ++i) {
			n80lx.text[7][7 + i] = sjis[i];
		}
		qsort(items, contents, sizeof(ITEM), item_sort);
		while (1) {
			int select_index = menu_select(items, contents);
			int iscurrent = 0;
			if (select_index >= 0) {
				switch (items[select_index].type) {
					case SOME_DIR:
						*isdir = 1;
						break;
					case N80_FILE:
					case CMT_FILE:
					case T88_FILE:
					case D88_FILE:
						*isdir = 0;
						*type = items[select_index].type;
						break;
					default:
						iscurrent = 1;
						break;
				}
				if (iscurrent)
					continue;
				memset(utf_buff, 0, 1024);
				memcpy(utf_buff, items[select_index].name, strlen(items[select_index].name));
				result = utf_buff;
			}
			break;
		}
		for (i = 0; i < contents; ++i)
			free(items[i].name);
	}
	free(items);
	return result;
}
/**--------------------------------------------------------------------------
 **		read N80 format file
 **------------------------------------------------------------------------*/
static BOOL read_n80file(const char *path) {
	BOOL		result = FALSE;
	uint16_t	spval;

	FILE *fp = fopen(path, "rb");
	if (fp != NULL) {
		fseek(fp, 0, SEEK_END);
		long fsiz = ftell(fp);
		rewind(fp);
		if (fsiz >= STDSIZE) {
			fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
			size_t rsiz = fread(n80device.mem._ram, 1, fsiz, fp);
			if (rsiz == fsiz) {
				fseek(fp, SP_VAL, SEEK_SET);
				rsiz = fread(&spval, 2, 1, fp);
				n80device.cpu.r.x.PC = PC_VAL;
				n80device.cpu.r.x.SP = spval;
				result = TRUE;
			}
		} else
			fprintf(stderr, "N80 file size error: %s\n", path);
		fclose(fp);
	} else
		fprintf(stderr, "can not open N80 file: %s\n", path);
	return result;
}
/**--------------------------------------------------------------------------
 **		read CMT format file
 **------------------------------------------------------------------------*/
static BOOL read_cmtfile(const char *path) {
	BOOL result = FALSE;
	FILE *fp = fopen(path, "rb");
	if (fp != NULL) {
		fseek(fp, 0, SEEK_END);
		long fsiz = ftell(fp);
		rewind(fp);
		if (fsiz < 64*1024) {
			long rsiz = (long)fread(cmt_buff, 1, (1024*64), fp);
			if (fsiz == rsiz) {
				n80device.cmt._size = (DWORD)rsiz;
				n80device.cmt._buff = cmt_buff;
				n80device.cmt._index = 0;
				result = TRUE;
			} else
				fprintf(stderr, "read CMT file size error: %s\n", path);
		} else
			fprintf(stderr, "too large CMT file size: %s\n", path);
		fclose(fp);
	} else
		fprintf(stderr, "can not open CMT file: %s\n", path);
	return result;
}
/**--------------------------------------------------------------------------
 **		read T88 format file
 **------------------------------------------------------------------------*/
static BOOL read_t88file(const char *path) {
	const char* header = "PC-8801 Tape Image(T88)";
	char buff[23 + 1];
	int  index = 0;
	BOOL result = false;

	FILE *fp = fopen(path, "rb");
	if (fp != NULL) {
		while (true) {
			// file header check
			if (fread(buff, 1, 24, fp) != 24) {
				fprintf(stderr, "header read error T88 file: %s\n", path);
				break;
			}
			if (strcmp(header, buff) != 0) {
				fprintf(stderr, "wrong header T88 file: %s\n", path);
				break;
			}
			// skip version tag
			if (fread(buff, 1, 6, fp) != 6) {
				fprintf(stderr, "version read error T88 file: %s\n", path);
				break;
			}
			while (true) {
				// tag + siz read
				if (fread(buff, 1, 4, fp) != 4) {
					fprintf(stderr, "tag read error T88 file: %s\n", path);
					break;
				}
				int tag_siz = 0;
				tag_siz += ((BYTE)buff[3]);
				tag_siz <<= 8;
				tag_siz += ((BYTE)buff[2]);
				// blank/space/mark tags
				if ((buff[0] == 0x00 && buff[1] == 0x01) ||
					(buff[0] == 0x02 && buff[1] == 0x01) ||
					(buff[0] == 0x03 && buff[1] == 0x01)) {
					if (fread(buff, 1, tag_siz, fp) != tag_siz) {
						fprintf(stderr, "tag size error T88 file: %s\n", path);
						break;
					}
				// data tag
				} else if (buff[0] == 0x01 && buff[1] == 0x01) {
					// area time/area length/data length/data type
					if (fread(buff, 1, 12, fp) != 12) {
						fprintf(stderr, "area size error T88 file: %s\n", path);
						break;
					}
					int data_siz = 0;
					data_siz += ((BYTE)buff[9]);
					data_siz <<= 8;
					data_siz += ((BYTE)buff[8]);
					// check buffer size
					if ((index + data_siz) > 64*1024) {
						fprintf(stderr, "data size overflow error T88 file: %s\n", path);
						break;
					}
					// read data
					if (fread(&cmt_buff[index], 1, data_siz, fp) != data_siz) {
						fprintf(stderr, "read data size error T88 file: %s\n", path);
						break;
					}
					index += data_siz;
				// end tag
				} else if (buff[0] == 0x00 && buff[1] == 0x00) {
					n80device.cmt._size = index;
					n80device.cmt._buff = cmt_buff;
					n80device.cmt._index = 0;
					result = TRUE;
					break;
				// invalid tag
				} else {
					fprintf(stderr, "invalid tag detect error T88 file: %s\n", path);
					break;
				}
			}
			break;
		}
		fclose(fp);
	} else
		fprintf(stderr, "can not open T88 file: %s\n", path);
	return result;
}
/**--------------------------------------------------------------------------
 **		read D88 format file
 **------------------------------------------------------------------------*/
static void d88_image_release(LPDISKIMAGE image) {
	if (--image->ref_count <= 0) {
		image->ref_count = 0;
		if (image->image_raw_data != NULL)
			free(image->image_raw_data);
		for (int i = 0; i < image->disk_count; ++i) {
			if (image->disks[i] != NULL) {
				free(image->disks[i]);
				image->disks[i] = NULL;
			}
		}
		free(image);
	}
}
static LPFLOPPYDISK d88_image_alloc_disk() {
	return (LPFLOPPYDISK)calloc(1, sizeof(FLOPPYDISK));
}
static BOOL read_d88file(const char *path, int slot_no, int unit_no) {
	BOOL result			= FALSE;
	LPDISKIMGMGR mgr	= &n80device.fdu.mgr;

	FILE *fp = fopen(path, "rb");
	if (fp == NULL)
		return FALSE;

	LPDISKIMAGE image = (LPDISKIMAGE)calloc(1, sizeof(DISKIMAGE));
	if (image != NULL) {
		fseek(fp, 0, SEEK_END);
		long fsiz = ftell(fp);
		rewind(fp);
		image->image_raw_data = (LPBYTE)calloc(1, fsiz);
		if (image->image_raw_data != NULL) {
			DWORD rsiz = fread(image->image_raw_data, 1, fsiz, fp);
			if (fsiz == rsiz) {
				image->image_length = rsiz;
				image->dispose		= d88_image_release;
				image->alloc		= d88_image_alloc_disk;
				result = mgr->attach(mgr, image, slot_no, unit_no);
			}
		}
		if (!result)
			d88_image_release(image);
	}
	fclose(fp);
	return result;
}
/**--------------------------------------------------------------------------
 **	main function
 **------------------------------------------------------------------------*/
static void chooser(LPCHOOSER this, int slot_no, int unit_no) {
	const char	*path = NULL;
	int			dir = 0, type = 0, i;

	for (i = 0; i < 25; ++i) {
		int j;
		for (j = 0; j < 80; ++j) {
			n80lx.text[i][j] = 32;
			n80lx.attr[i][j] = (i == 4 || i == 5) && (j > 0 && j < 79) ? 0x27 : i == 7 && (j > 0 && j < 79) ? 0x26 : 0x23;
		}
	}
	const char *title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for Linux (SDL2)";
	const char *title2 = "    Copyright 1993-2021 (C) H.Kanemoto, All Right Reserved.";
	const char *title3 = "Image File Chooser (CMT/N80/T88/D88)";
	const char *title4 = "usage:  [ESC] Quit,  [RET] Select, [\x1e\x1f\x1d\x1c] Move, [PgUp/PgDn] Page";
	const char *title5 = "dir: ";
	memcpy(&n80lx.text[1][2], title1, strlen(title1));
	memcpy(&n80lx.text[2][2], title2, strlen(title2));
	memcpy(&n80lx.text[4][2], title3, strlen(title3));
	memcpy(&n80lx.text[5][2], title4, strlen(title4));
	memcpy(&n80lx.text[7][2], title5, strlen(title5));

	// 前回開いたdirを復元
	DIR *d = opendir(this->dirpath);
	if (d == NULL) {
		// ダメならカレントdirを開く
		strcpy(this->dirpath, n80lx.home);
		d = opendir(this->dirpath);
		if (d == NULL)
			return;
	}
	closedir(d);
	if (chdir(this->dirpath) == 0) {
		while ((path = domenu(this->dirpath, slot_no != -1, &dir, &type)) != NULL) {
			if (dir) {			// 選択されたdirに移動
				if (chdir(path) == 0) {
					char *dp = getcwd(this->dirpath, sizeof(this->dirpath));
					if (dp == NULL)
						break;
				}
			} else {			// 選択されたファイルを読み込む
				switch (type) {
				case N80_FILE:	// N80形式ファイル
					read_n80file(path);			break;
				case CMT_FILE:	// CMT形式ファイル
					read_cmtfile(path);			break;
				case T88_FILE:	// T88形式ファイル
					read_t88file(path);			break;
				case D88_FILE:	// D88形式ファイル
					read_d88file(path, slot_no == -1 ? 0 : slot_no, unit_no);
					break;
				}
				break;
			}
		}
		n80lx.ini.flush(&n80lx.ini);	// save corrent opened dir
	}
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void init_chooser(LPCHOOSER this) {
	this->open			= chooser;
	this->read_cmtfile	= read_cmtfile;
	this->read_n80file	= read_n80file;
	this->read_t88file	= read_t88file;
	this->read_d88file	= read_d88file;
	this->utf8to932		= utf8to932;
}
