/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2021 H.Kanemoto
 **
 **		json file access
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h> 
#include <cjson/cJSON.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#include <linux/limits.h>
#define	export
/**--------------------------------------------------------------------------
 **	external variable/functions
 **------------------------------------------------------------------------*/
extern N80LX		n80lx;
extern N80DEVICE	n80device;
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static char path[PATH_MAX];
/**--------------------------------------------------------------------------
 **	local functions
 **------------------------------------------------------------------------*/
static BOOL ini_create_default(PINI self) {
	n80device.param.interlace = TRUE;
	n80device.vrtc.crt_hires_mode = _CRT_NORMAL;
	n80lx.tenkeyless = 0;
	char *dp = getcwd(n80lx.chooser.dirpath, sizeof(n80lx.chooser.dirpath));
	if (dp == NULL)
		strcpy(n80lx.chooser.dirpath, ".");
	strcpy(n80lx.n80rom_filename, "PC-8001.ROM");
	strcpy(n80lx.fdurom_filename, "pdisk.rom");
	strcpy(n80lx.bdf_filename, "shnmk16.bdf");
	strcpy(n80lx.snd_dev_name, "");
	return self->flush(self);
}
static int ini_purchase_intvalue(cJSON *json, const char *valuename) {
	cJSON *item = cJSON_GetObjectItemCaseSensitive(json, valuename);
	if (item != NULL)
		return item->valueint;
	else
		return -1;
}
static void ini_purchase_strvalue(cJSON *json, const char *valuename, int maxlen, char *buffer) {
	cJSON *item = cJSON_GetObjectItemCaseSensitive(json, valuename);
	if (item != NULL && cJSON_IsString(item) && item->valuestring != NULL && strlen(item->valuestring) < maxlen)
		strcpy(buffer, item->valuestring);
	else
		strcpy(buffer, "");
}
/**--------------------------------------------------------------------------
 **	local functions (methods)
 **------------------------------------------------------------------------*/
static BOOL ini_load(PINI self) {
	BOOL result = FALSE;
	FILE *fp = fopen(path, "r");
	if (fp != NULL) {
		LPVOID buffer = NULL;
		fseek(fp, 0, SEEK_END);
		int fsiz = ftell(fp);
		rewind(fp);
		buffer = malloc(fsiz);
		if (buffer != NULL) {
			size_t rsiz = fread(buffer, 1, fsiz, fp);
			if (rsiz == fsiz) {
				cJSON *json = cJSON_Parse((const char *)buffer);
				if (json != NULL) {
					n80device.param.interlace = ini_purchase_intvalue(json, "interlace") == 1;
					n80device.vrtc.crt_hires_mode = ini_purchase_intvalue(json, "hirescrt") == 1 ? _CRT_HIRES : _CRT_NORMAL;
					n80lx.tenkeyless = ini_purchase_intvalue(json, "tenkeyless") == 1;
					ini_purchase_strvalue(json, "n80rom", 64, n80lx.n80rom_filename);
					ini_purchase_strvalue(json, "fdurom", 64, n80lx.fdurom_filename);
					ini_purchase_strvalue(json, "chooser", PATH_MAX, n80lx.chooser.dirpath);
					ini_purchase_strvalue(json, "snddev", 64, n80lx.snd_dev_name);
					ini_purchase_strvalue(json, "bdf", 64, n80lx.bdf_filename);

					cJSON_Delete(json);
					result = TRUE;
				}
			}
			free(buffer);
		}
		fclose(fp);
	}
	return result;
}
static BOOL ini_flush(PINI self) {
	BOOL result = FALSE;
	cJSON *json = cJSON_CreateObject();
	cJSON *interlace = n80device.param.interlace ? cJSON_CreateTrue() : cJSON_CreateFalse();
	cJSON *hirescrt  = n80device.vrtc.crt_hires_mode == _CRT_HIRES  ? cJSON_CreateTrue() : cJSON_CreateFalse();
	cJSON *tenkeyless  = n80lx.tenkeyless == 1  ? cJSON_CreateTrue() : cJSON_CreateFalse();
	cJSON_AddItemToObject(json, "interlace", interlace);
	cJSON_AddItemToObject(json, "hirescrt", hirescrt);
	cJSON_AddItemToObject(json, "tenkeyless", tenkeyless);
	cJSON_AddStringToObject(json, "n80rom", n80lx.n80rom_filename);
	cJSON_AddStringToObject(json, "fdurom", n80lx.fdurom_filename);
	cJSON_AddStringToObject(json, "chooser", n80lx.chooser.dirpath);
	cJSON_AddStringToObject(json, "bdf", n80lx.bdf_filename);
	cJSON_AddStringToObject(json, "snddev", n80lx.snd_dev_name);

	char *json_str = cJSON_Print(json);

	FILE *fp = fopen(path, "w");
	if (fp != NULL) {
		fputs(json_str, fp);
		fclose(fp);
		result = TRUE;
	}
	cJSON_free(json_str);
	cJSON_Delete(json);
	return result;
}
static void ini_dispose(PINI self) {}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void init_json(PINI self) {
	self->init		= init_json;
	self->dispose	= ini_dispose;
	self->flush		= ini_flush;
	self->load		= ini_load;

	memset(path, 0, sizeof(path));
	strncpy(path, n80lx.home, sizeof(n80lx.home));
	strcat(path, "/n80.json");
	fprintf(stderr, "json path = %s\n", path);

	if (!self->load(self))
		ini_create_default(self);
}
