/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		keyboard
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#define	export
/**--------------------------------------------------------------------------
 **	external variable
 **------------------------------------------------------------------------*/
extern N80LX	n80lx;
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static BYTE  down_matorix[80][2] = {
	{0,0xfe}, {0,0xfd}, {0,0xfb}, {0,0xf7},
	{0,0xef}, {0,0xdf}, {0,0xbf}, {0,0x7f},
	{1,0xfe}, {1,0xfd}, {1,0xfb}, {1,0xf7},
	{1,0xef}, {1,0xdf}, {1,0xbf}, {1,0x7f},
	{2,0xfe}, {2,0xfd}, {2,0xfb}, {2,0xf7},
	{2,0xef}, {2,0xdf}, {2,0xbf}, {2,0x7f},
	{3,0xfe}, {3,0xfd}, {3,0xfb}, {3,0xf7},
	{3,0xef}, {3,0xdf}, {3,0xbf}, {3,0x7f},
	{4,0xfe}, {4,0xfd}, {4,0xfb}, {4,0xf7},
	{4,0xef}, {4,0xdf}, {4,0xbf}, {4,0x7f},
	{5,0xfe}, {5,0xfd}, {5,0xfb}, {5,0xf7},
	{5,0xef}, {5,0xdf}, {5,0xbf}, {5,0x7f},
	{6,0xfe}, {6,0xfd}, {6,0xfb}, {6,0xf7},
	{6,0xef}, {6,0xdf}, {6,0xbf}, {6,0x7f},
	{7,0xfe}, {7,0xfd}, {7,0xfb}, {7,0xf7},
	{7,0xef}, {7,0xdf}, {7,0xbf}, {7,0x7f},
	{8,0xfe}, {8,0xfd}, {8,0xfb}, {8,0xf7},
	{8,0xef}, {8,0xdf}, {8,0xbf}, {8,0x7f},
	{9,0xfe}, {9,0xfd}, {9,0xfb}, {9,0xf7},
	{9,0xef}, {9,0xdf}, {9,0xbf}, {9,0x7f}
};
static BYTE  up_matorix[80][2] = {
	{0,0x01}, {0,0x02}, {0,0x04}, {0,0x08},
	{0,0x10}, {0,0x20}, {0,0x40}, {0,0x80},
	{1,0x01}, {1,0x02}, {1,0x04}, {1,0x08},
	{1,0x10}, {1,0x20}, {1,0x40}, {1,0x80},
	{2,0x01}, {2,0x02}, {2,0x04}, {2,0x08},
	{2,0x10}, {2,0x20}, {2,0x40}, {2,0x80},
	{3,0x01}, {3,0x02}, {3,0x04}, {3,0x08},
	{3,0x10}, {3,0x20}, {3,0x40}, {3,0x80},
	{4,0x01}, {4,0x02}, {4,0x04}, {4,0x08},
	{4,0x10}, {4,0x20}, {4,0x40}, {4,0x80},
	{5,0x01}, {5,0x02}, {5,0x04}, {5,0x08},
	{5,0x10}, {5,0x20}, {5,0x40}, {5,0x80},
	{6,0x01}, {6,0x02}, {6,0x04}, {6,0x08},
	{6,0x10}, {6,0x20}, {6,0x40}, {6,0x80},
	{7,0x01}, {7,0x02}, {7,0x04}, {7,0x08},
	{7,0x10}, {7,0x20}, {7,0x40}, {7,0x80},
	{8,0x01}, {8,0x02}, {8,0x04}, {8,0x08},
	{8,0x10}, {8,0x20}, {8,0x40}, {8,0x80},
	{9,0x01}, {9,0x02}, {9,0x04}, {9,0x08},
	{9,0x10}, {9,0x20}, {9,0x40}, {9,0x80}
};
/**--------------------------------------------------------------------------
 **	local function
 **------------------------------------------------------------------------*/
static inline BYTE convert_key(SDL_Keycode code) {
	BYTE	key = 0xff;
	switch (code) {
	case SDLK_1:	/* 1 */	key = 0x31;	break;
	case SDLK_2:	/* 2 */	key = 0x32;	break;
	case SDLK_3:	/* 3 */	key = 0x33;	break;
	case SDLK_4:	/* 4 */	key = 0x34;	break;
	case SDLK_5:	/* 5 */	key = 0x35;	break;
	case SDLK_6:	/* 6 */	key = 0x36;	break;
	case SDLK_7:	/* 7 */	key = 0x37;	break;
	case SDLK_8:	/* 8 */	key = 0x38;	break;
	case SDLK_9:	/* 9 */	key = 0x39;	break;
	case SDLK_0:	/* 0 */	key = 0x30;	break;

	case SDLK_q:	/* Q */	key = 0x21;	break;
	case SDLK_w:	/* W */	key = 0x27;	break;
	case SDLK_e:	/* E */	key = 0x15;	break;
	case SDLK_r:	/* R */	key = 0x22;	break;
	case SDLK_t:	/* T */	key = 0x24;	break;
	case SDLK_y:	/* Y */	key = 0x29;	break;
	case SDLK_u:	/* U */	key = 0x25;	break;
	case SDLK_i:	/* I */	key = 0x19;	break;
	case SDLK_o:	/* O */	key = 0x1F;	break;
	case SDLK_p:	/* P */	key = 0x20;	break;

	case SDLK_a:	/* A */	key = 0x11;	break;
	case SDLK_s:	/* S */	key = 0x23;	break;
	case SDLK_d:	/* D */	key = 0x14;	break;
	case SDLK_f:	/* F */	key = 0x16;	break;
	case SDLK_g:	/* G */	key = 0x17;	break;
	case SDLK_h:	/* H */	key = 0x18;	break;
	case SDLK_j:	/* J */	key = 0x1A;	break;
	case SDLK_k:	/* K */	key = 0x1B;	break;
	case SDLK_l:	/* L */	key = 0x1C;	break;

	case SDLK_z:	/* Z */	key = 0x2A;	break;
	case SDLK_x:	/* X */	key = 0x28;	break;
	case SDLK_c:	/* C */	key = 0x13;	break;
	case SDLK_v:	/* V */	key = 0x26;	break;
	case SDLK_b:	/* B */	key = 0x12;	break;
	case SDLK_n:	/* N */	key = 0x1E;	break;
	case SDLK_m:	/* M */	key = 0x1D;	break;

	case SDLK_ESCAPE:		/* ESC   */	key = 0x4F;	break;
	case SDLK_MINUS:		/* = -   */	key = 0x2F;	break;
	case SDLK_CARET:		/* ^ ~   */	key = 0x2E;	break;
	case SDLK_BACKSPACE:	/* BS    */	key = 0x48;	break;
//	case 0x0F:				/* TAB	 */	key = 0x50;	break;
	case SDLK_AT:	/* @     */	key = 0x10;	break;
	case SDLK_LEFTBRACKET:	/* [     */	key = 0x2B;	break;
	case SDLK_RETURN:		/* ENTER */	key = 0x0F;	break;
	case SDLK_CAPSLOCK:
	case SDLK_LCTRL:		/* CTL   */	key = 0x47;	break;
	case SDLK_SEMICOLON:	/* + ;   */	key = 0x3B;	break;
	case SDLK_COLON:		/* * :   */	key = 0x3A;	break;
	case SDLK_LSHIFT:		/* L-SHF */	key = 0x46;	break;
	case SDLK_RIGHTBRACKET:	/* ]     */	key = 0x2D;	break;
	case SDLK_COMMA:		/* < ,   */	key = 0x3C;	break;
	case SDLK_PERIOD:		/* > .   */	key = 0x3D;	break;
	case SDLK_SLASH:		/* / ?   */	key = 0x3E;	break;
	case SDLK_RSHIFT:		/* R-SHF */	key = 0x46;	break;
	case SDLK_ASTERISK:		/* *     */	key = 0x0A;	break;
	case 1073740963:		/* GRAPH */	key = 0x44;	break;
	case SDLK_LALT:			/* GRAPH */	key = 0x44;	break;
	case SDLK_SPACE:		/* SP    */	key = 0x4E;	break;
	case SDLK_F1:			/* F1    */	key = 0x49;	break;
	case SDLK_F2:			/* F2    */	key = 0x4A;	break;
	case SDLK_F3:			/* F3    */	key = 0x4B;	break;
	case SDLK_F4:			/* F4    */	key = 0x4C;	break;
	case SDLK_F5:			/* F5    */	key = 0x4D;	break;
//	case 0x3E:	/* CAN   */	key = n80lx.fn_on ? 0x48 : 0x4F;	break;	// STOP/ESC
	case SDLK_KP_7:			/* TEN7  */	key = 0x07;	break;
	case SDLK_KP_8:			/* TEN8  */	key = 0x08;	break;
	case SDLK_KP_9:			/* TEN9  */	key = 0x09;	break;
	case SDLK_KP_4:			/* TEN4  */	key = 0x04;	break;
	case SDLK_KP_5:			/* TEN5  */	key = 0x05;	break;
	case SDLK_KP_6:			/* TEN6  */	key = 0x06;	break;
	case SDLK_KP_PLUS:		/* TEN+  */	key = 0x0B;	break;
	case SDLK_KP_1:			/* TEN1  */	key = 0x01;	break;
	case SDLK_KP_2:			/* TEN2  */	key = 0x02;	break;
	case SDLK_KP_3:			/* TEN3  */	key = 0x03;	break;
	case SDLK_KP_0:			/* TEN0  */	key = 0x00;	break;
	case SDLK_KP_PERIOD:	/* TEN.  */	key = 0x0E;	break;
	case SDLK_BACKSLASH:	/* _     */	key = 0x3F;	break;
	case SDLK_RALT:			/* KANA  */	key = 0x45;	break;
	case SDLK_HOME:			/* HOME  */	key = 0x40;	break;
	case SDLK_INSERT:		/* INS   */	key = 0x43;	break;
	case SDLK_DELETE:		/* DEL   */	key = 0x43;	break;
	case 165:				/* \ |   */	key = 0x2C;	break;
	case SDLK_UP:			/* ARW-U */	key = !n80lx.tenkeyless ? 0x41 : 0x08;	break;
	case SDLK_LEFT:			/* ARW-L */	key = !n80lx.tenkeyless ? 0x42 : 0x04;	break;
	case SDLK_RIGHT:		/* ARW-R */	key = !n80lx.tenkeyless ? 0x42 : 0x06;	break;
	case SDLK_DOWN:			/* ARR-D */	key = !n80lx.tenkeyless ? 0x41 : 0x02;	break;
	default:	break;
	}
	return key;
}
/**--------------------------------------------------------------------------
 **	global functions
 **------------------------------------------------------------------------*/
export void onkey_down(SDL_Keycode code) {
	BYTE	key = convert_key(code);

	if (key != 0xff) {
		int  index = down_matorix[key][0];
		BYTE bit   = down_matorix[key][1];
		n80device.io.keys[index] &= bit;
	}
}

export void onkey_up(SDL_Keycode code) {
	BYTE	key = convert_key(code);

	if (key != 0xff) {
		int  index = up_matorix[key][0];
		BYTE bit   = up_matorix[key][1];
		n80device.io.keys[index] |= bit;
	}
}
