/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		main
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <locale.h>
#include <sys/ioctl.h>
//#include <sys/mman.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#include <signal.h>
#include <sys/timerfd.h>
//#include <linux/time.h>
#define	export
/**--------------------------------------------------------------------------
 **	external variables/functions
 **------------------------------------------------------------------------*/
export N80LX	n80lx;
extern int		init_font();						/** sdl2.c		**/
extern void		enter(void);
extern void		leave(void);
extern void		onkey_down(SDL_Keycode);			/** key.c		**/
extern void		onkey_up(SDL_Keycode);
extern void		init_chooser(LPCHOOSER);			/** file.c		**/
extern void		init_config(LPCONFIG);				/** config.c	**/
extern int		init_alsa(void);					/** alsa.c		**/
extern void		dispose_alsa();
extern void		init_json(PINI);					/** json.c		**/
extern bool		init_bdf(PBDF, const char *);		/** bdf.c		**/
extern void		about(void);						/** about.c		**/
extern void		_n80_device_initialize(LPN80DEVICE, SYNCFNC enter, SYNCFNC leave);
/**--------------------------------------------------------------------------
 **	timer thread
 **------------------------------------------------------------------------*/
static volatile int timer_quit = 0;
static volatile int timer_pause = 0;
static void sig_abort_func(int sig) {
	fprintf(stderr, "sigint abort\n");
	timer_quit = 1;
}
static void timer_thread_func(void) {
	int tfd;
	struct itimerspec stime;
	ssize_t r_siz;
	u_int64_t cnt;

	tfd = timerfd_create(CLOCK_MONOTONIC, 0);
	stime.it_value.tv_sec  = 0;
    stime.it_value.tv_nsec = 16666666;
    stime.it_interval.tv_sec  = 0;
    stime.it_interval.tv_nsec = 16666666;
    timerfd_settime(tfd, 0, &stime, NULL);
	fprintf(stderr, "timerfd start\n");
	do {
		r_siz = read(tfd, &cnt, sizeof(u_int64_t));
		if (r_siz < 0)
			continue;
		if (timer_quit)
			break;
		if (!timer_pause) {
			SDL_Event event;
			SDL_memset(&event, 0, sizeof(event));
			event.type = SDL_USEREVENT;
			event.user.code = n80lx.pipe_evt;
			SDL_PushEvent(&event);
		}
	} while (1);
	fprintf(stderr, "timerfd close\n");
	close(tfd);
}
/**--------------------------------------------------------------------------
 **	Z80 thread
 **------------------------------------------------------------------------*/
static void *z80_thread_func(void *param) {
	volatile double	est_time = n80device.get_tick() + n80device.param.interval_sec;
	volatile DWORD	state = 0, req;
	volatile int	quit = 0;
	long 			sync_T = 0;	// main <=> sub cpu state adjust count
#if _VERBOSE
	volatile int	delay_count = 0;
#endif

	for (UINT tk = 0; !quit; ++tk) {
		// exit z80 thread if detect cpu request
		req = interlocked_exchange(&n80lx.request_cpu_signal, 0);
		if (req == 0)
			;
		else if (req == 1) {
			fprintf(stderr, "ACCEPT REQUEST CPU STOP\n");
			break;
		}
		else if (req == 2) {	// accept reset
			fprintf(stderr, "ACCEPT RESET CPU STOP\n");
			n80device.reset(&n80device);
			n80device.cpu.reset(&n80device.cpu);
			n80device.attr.invalid();
			n80device.snd.beep_off();
			n80device.snd.sound_off();
			n80device.snd.clear_ticks();
			est_time  = n80device.get_tick() + n80device.param.interval_sec;
			state     = 0;
			n80device.dmac._penarty = 0;
		}
		else if (req == 3) {	// accept pause
			timer_pause = 1;
			fprintf(stderr, "ACCEPT PAUSE CPU STOP\n");
			pthread_mutex_lock(&n80lx.pause_sync);
			pthread_mutex_unlock(&n80lx.pause_sync);
			timer_pause = 0;
			est_time  = n80device.get_tick() + n80device.param.interval_sec;
			//state     = 0;
			//n80device.dmac._penarty = 0;
		}
		n80device.cpu.r.x.T		= 0;
		n80device.cpu.r.x.TwM1	= 0;
		n80device.cpu.r.x.W		= 0;
		n80device.cpu.r.x.Wh	= 0;
		sync_T					= 0;

		// z80 step execute
		if (n80device.dmac._penarty > 0) {
			if (n80device.dmac._penarty > 2) {
				n80device.cpu.r.x.T = 2;
				n80device.dmac._penarty -= 2;
			} else {
				n80device.cpu.r.x.T = n80device.dmac._penarty;
				n80device.dmac._penarty = 0;
			}
		} else {
			n80device.cpu.extrn_sync_clk = tk;
			n80device.cpu.step(&n80device.cpu);
			// Z80 step execute
			if (n80device.cpu.halt_detect && n80device.cpu.r.x.HALT != _Z80HALT_76) {
				// notify if detect illegal instructions
				fprintf(stderr, "HALT\n");
				quit = 1;
				break;
			}
			sync_T = n80device.cpu.r.x.T + n80device.cpu.r.x.TwM1;
		}
		n80device.cpu.r.x.T += (n80device.cpu.r.x.W + n80device.cpu.r.x.Wh + n80device.cpu.r.x.TwM1);

		// SUB CPU STEP実行
		n80device.fdu.cpu.extrn_sync_clk = tk;
		n80device.fdu.cpu_sync_T += sync_T;
		if (n80device.fdu.cpu_sync_T > 0)
			n80device.fdu.cpu.step(&n80device.fdu.cpu);

		// vrtc operation
		int vrtc = n80device.vrtc.blanking(&n80device.cpu);
		if (vrtc == _VRTC_VBBGN) {
			;
		} else if (vrtc == _VRTC_VBEND) {
			// draw screen
			//SDL_Event event;
			//SDL_memset(&event, 0, sizeof(event));
			//event.type = SDL_USEREVENT;
			//event.user.code = n80lx.pipe_evt;
			//SDL_PushEvent(&event);
		}
		// give peripherals a chance to process
		n80device.dispatch(&n80device);
		// beep/pcg ticks
		n80device.snd.append_ticks(n80device.cpu.r.x.T);
		
		interlocked_exchangeadd(&n80lx.sys_Ts, n80device.cpu.r.x.T);

		// cpu clock sync
		state += n80device.cpu.r.x.T;
		if (state >= n80device.param.sync_ticks) {
			// let alsa play sound
			int write_frames = n80device.snd.make_pcm();
			n80device.snd.feed_pcm(write_frames);

			volatile double cur_time = n80device.get_tick();
			while (cur_time < est_time) {
				volatile double diff = est_time - cur_time;
				if (diff >= 0.0015) {		// 1.5ms以上ある場合はsleep
					struct timespec req;
					req.tv_sec = 0;
					req.tv_nsec = (diff - 0.0005) * 1e9;	// ﾏｰｼﾞﾝで500us分差し引く
					nanosleep(&req, NULL);
					//clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &req, NULL);
				} else if (diff < 0.001)	// 1ms以下なら抜ける
					break;
				cur_time = n80device.get_tick();
			}
			est_time += n80device.param.interval_sec;
			cur_time = n80device.get_tick();
			if (cur_time >= est_time) {
				est_time = cur_time + n80device.param.interval_sec;
#if _VERBOSE
				++delay_count;
#endif
			}
			state -= n80device.param.sync_ticks;
#if _VERBOSE
			if ((tk % 100) == 0) {
				if (delay_count > 0) {
					fprintf(stderr, "delay ave = %.3f(%%)\n", ((double)delay_count / 100.0) * 100);
				}
				delay_count = 0;
			}
#endif
		}
	}
	return NULL;
}
/**--------------------------------------------------------------------------
 **	N80 main thread
 **------------------------------------------------------------------------*/
int eventFilter(void *userdata, SDL_Event *event) {
	if (event->type == SDL_KEYDOWN || event->type == SDL_KEYUP || 
		event->type == SDL_USEREVENT || event->type == SDL_QUIT) {
		return 1; // イベントを許可
	} else {
		return 0; // イベントをブロック
	}
}
static inline void clear_special_key(void) {
	n80lx.shift = false;
	n80lx.alt   = false;
	n80lx.ctl   = false;
	n80lx.ins	= false;
	n80lx.del	= false;
}
static inline void enter_critical(void) {
	// z80, alsa threadを一時停止
	pthread_mutex_lock(&n80lx.pause_sync);
	interlocked_exchange(&n80lx.request_cpu_signal, 3);
	n80lx.save_pcg = n80device.pcg_mode;
	n80device.pcg_mode = 0;
}
static inline void leave_critical(void) {
	n80device.pcg_mode = n80lx.save_pcg;
	n80device.attr.invalid();
	n80device.io.clear_keys();
	clear_special_key();
	#ifdef _ZAURUS_
	clr_fb_display(n80lx.video);
	#endif
	pthread_mutex_unlock(&n80lx.pause_sync);
}
static void n80main(void) {
	int			cnt = 0;//, kcnt = 0;
	double		prev = n80device.get_tick(), interval = 0;
	SDL_Event	e;

	if (signal(SIGINT, sig_abort_func) == SIG_ERR) {
		fprintf(stderr, "can not regist sigint\n");
		return;
	}
	pthread_t th_t;
	int ret = pthread_create(&th_t, NULL, (void *)timer_thread_func, NULL);
	if (ret != 0) {
		fprintf(stderr, "can not create timer thread\n");
		return;
	}
	clear_special_key();
	// 100msのタイムアウト付きでキー入力監視
	SDL_AddEventWatch(eventFilter, NULL);
	for (int quit = 0; !quit;) {
		while (SDL_WaitEvent(&e) != 0) {
			if (e.type == SDL_QUIT) {
				quit = 1;
				break;
			} else if (e.type == SDL_KEYDOWN) {
				if (e.key.repeat == 0) {
					//fprintf(stderr, "keyup ↓code=%d  cont=%d       \r", e.key.keysym.sym,  ++kcnt);
					if (e.key.keysym.sym == SDLK_LSHIFT || e.key.keysym.sym == SDLK_RSHIFT)
						n80lx.shift = true;
					else if (e.key.keysym.sym == SDLK_LALT || e.key.keysym.sym == SDLK_RALT)
						n80lx.alt = true;
					else if (e.key.keysym.sym == SDLK_LCTRL || e.key.keysym.sym == SDLK_RCTRL)
						n80lx.ctl = true;
					else if (e.key.keysym.sym == SDLK_DELETE)
						n80lx.del = true;
					else if (e.key.keysym.sym == SDLK_INSERT)
						n80lx.ins = true;
					else if (n80lx.alt && n80lx.ins && (e.key.keysym.sym == SDLK_1 || e.key.keysym.sym == SDLK_KP_1)) {
						enter_critical();
						n80lx.chooser.open(&n80lx.chooser, 0, -1);
						leave_critical();
						continue;
					} else if (n80lx.alt && n80lx.ins && (e.key.keysym.sym == SDLK_2 || e.key.keysym.sym == SDLK_KP_2)) {
						enter_critical();
						n80lx.chooser.open(&n80lx.chooser, 1, -1);
						leave_critical();
						continue;
					} else if (n80lx.alt && n80lx.del && (e.key.keysym.sym == SDLK_1 || e.key.keysym.sym == SDLK_KP_1)) {
						enter_critical();
						n80device.fdu.mgr.detach(&n80device.fdu.mgr, 0);
						leave_critical();
						continue;
					} else if (n80lx.alt && n80lx.del && (e.key.keysym.sym == SDLK_2 || e.key.keysym.sym == SDLK_KP_2)) {
						enter_critical();
						n80device.fdu.mgr.detach(&n80device.fdu.mgr, 1);
						leave_critical();
						continue;
					}
					onkey_down(e.key.keysym.sym);
				}
			} else if (e.type == SDL_KEYUP) {
				//fprintf(stderr, "keyup ↑code=%d  cont=%d       \r", e.key.keysym.sym,  --kcnt);
				if (e.key.keysym.sym == SDLK_LSHIFT || e.key.keysym.sym == SDLK_RSHIFT)
					n80lx.shift = false;
				else if (e.key.keysym.sym == SDLK_LALT || e.key.keysym.sym == SDLK_RALT)
					n80lx.alt = false;
				else if (e.key.keysym.sym == SDLK_LCTRL || e.key.keysym.sym == SDLK_RCTRL)
					n80lx.ctl = false;
				else if (e.key.keysym.sym == SDLK_DELETE)
					n80lx.del = false;
				else if (e.key.keysym.sym == SDLK_INSERT)
					n80lx.ins = false;
				else if (e.key.keysym.sym == SDLK_END) {
					quit = 1;
					break;
				}
				switch (e.key.keysym.sym) {
					case SDLK_F12:	// [F12]Keyでリセット
						if (!n80lx.shift) {
							interlocked_exchange(&n80lx.request_cpu_signal, 2);
							n80device.pcg_mode = 0;
							n80device.cmt._index = 0;
							n80device.io.clear_keys();
							n80device.snd.sound_off();
						} else {
							if (SDL_GetWindowFlags(n80lx.window) & SDL_WINDOW_FULLSCREEN_DESKTOP)
								SDL_SetWindowFullscreen(n80lx.window, 0);
							else
								SDL_SetWindowFullscreen(n80lx.window, SDL_WINDOW_FULLSCREEN_DESKTOP);
							break;
						}
						break;
					case SDLK_F6:	// [F6]KeyでAbout/SysClk
						if (n80lx.shift) {
							n80lx.show_sys_clock = !n80lx.show_sys_clock;
							break;
						}
					case SDLK_F7:	// [F7]KeyでConfig
					case SDLK_F9: {	// [F9]KeyでFile Chooser
						enter_critical();
						if (e.key.keysym.sym == SDLK_F9)	
							n80lx.chooser.open(&n80lx.chooser, -1, -1);
						else if (e.key.keysym.sym == SDLK_F6)
							about();
						else
							n80lx.config.open(&n80lx.config);
						leave_critical();
						break;
					}
					case SDLK_F11:	// [F11]KeyでPCG Toggle
						// toggle PCG mode ON/OFF
						n80device.pcg_mode ^= 0x01;
						n80device.attr.invalid();
						break;
					default:
						onkey_up(e.key.keysym.sym);
						break;
				}
			} else if (e.type == SDL_USEREVENT && e.user.code == n80lx.pipe_evt) {
				// draw 8001 screen on texture
				n80device.blt.reflesh();
				// reflect texture on screen
				SDL_UpdateTexture(n80lx.texture, NULL, n80lx.video, 640 * 4);
				SDL_RenderClear(n80lx.renderer);
				SDL_RenderCopy(n80lx.renderer, n80lx.texture, NULL, NULL);
				SDL_RenderPresent(n80lx.renderer);
				SDL_FlushEvent(SDL_USEREVENT);

				if (n80lx.show_sys_clock) {
					double current = n80device.get_tick();
					interval += (current - prev);
					prev = current;
					++cnt;

					if (interval >= 0.5) {
						n80lx.flips = (int) floor(((double)cnt / (double) interval) + 0.5);

						DWORD cpuTs = interlocked_exchange(&n80device.cpu.Ts, 0);
						n80lx.eff_hz = ((double)cpuTs / (double)interval) / 1000000.0;
						n80lx.eff_hz = floor((n80lx.eff_hz + 0.00005) * 1000.0) / 1000.0;
						
						DWORD sysTs = interlocked_exchange(&n80lx.sys_Ts, 0);
						n80lx.sys_hz = ((double)sysTs / (double)interval) / 1000000.0;
						n80lx.sys_hz = floor((n80lx.sys_hz + 0.00005) * 1000.0) / 1000.0;

						char subcpu = n80device.fdu.cpu_sleeping == 1 ? '-' : 'A';
						char d1 = n80device.fdu.fdd.drives[0].get_state(&n80device.fdu.fdd.drives[0]);
						char d2 = n80device.fdu.fdd.drives[1].get_state(&n80device.fdu.fdd.drives[1]);
						char p1 = n80device.fdu.ppi.state1;
						char p2 = n80device.fdu.ppi.state2;
						char phase = '-';
						for (int d = 0; d < 2; ++d) {
							switch (n80device.fdu.fdc.phase) {
							case FDC_PHASE_CMD:		phase = 'C';	break;
							case FDC_PHASE_EXE:		phase = 'E';	break;
							case FDC_PHASE_RESULT:	phase = 'R';	break;
							}
						}
						n80device.fdu.ppi.state1 = '-';
						n80device.fdu.ppi.state2 = '-';
						fprintf(stderr, " CLK/CPU=[%.3f/%.3f](MHz) SUB[%c] FDC[%c] DRV[%c/%c] PPI[%c/%c] FPS[%d]\r",
							n80lx.sys_hz, n80lx.eff_hz, subcpu, phase, d1, d2, p1, p2, n80lx.flips);
						interval = 0;
						cnt = 0;
					}
				}
			}
		}
	}
	timer_quit = 1;
	pthread_join(th_t, NULL);
}
/**--------------------------------------------------------------------------
 **	処理分岐
 **------------------------------------------------------------------------*/
int branch(void) {
	// z80/disp threads
	pthread_t th;
	if (pthread_create(&th, NULL, z80_thread_func, NULL) != 0) {
		fprintf(stderr, "can not create z80 thread\n");
	} else {
		n80main();
		interlocked_exchange(&n80lx.request_cpu_signal, 1);

		pthread_join(th, NULL);
	}
	interlocked_exchange(&n80lx.request_cpu_signal, 1);
	return 0;
}
/**--------------------------------------------------------------------------
 **	主関数
 **------------------------------------------------------------------------*/
int main_core(void) {
	int		result = 1, rsiz, fsiz;
	FILE	*fp;
	
	// ロケール設定
	setlocale(LC_ALL, "");

	// 画面クリア、カーソル消去
	//printf("\x1b[2J\n");
	//printf("\x1b[>5h\n");

	// N80 structure clean
	memset(&n80lx, 0, sizeof(n80lx));

	// get OS version
	uname(&n80lx.utsname);

	// home dir
	char *dp = getcwd(n80lx.home, sizeof(n80lx.home));
	if (dp == NULL)
		strncpy(n80lx.home, ".", sizeof(n80lx.home) - 1);
	fprintf(stderr, "home path = %s\n", n80lx.home);

	// PC-8001 Device Initialize
	memset(&n80device, 0, sizeof(n80device));
	_n80_device_initialize(&n80device, enter, leave);

	// ini json
	init_json(&n80lx.ini);

	// ini config
	init_config(&n80lx.config);

	// init bdf
	init_bdf(&n80lx.bdf, n80lx.bdf_filename);

	// ini chooser
	init_chooser(&n80lx.chooser);

	n80device.blt.chg_palette(n80device.param.interlace);

	// font file load
	if (init_font())
		return 1;

	// PC-8001 rom file load
	fp = fopen(n80lx.n80rom_filename, "rb");
	if (fp == NULL) {
		fprintf(stderr, "can not open ROM file \"%s\"\n", n80lx.n80rom_filename);
		goto err_exit;
	}
	fseek(fp, 0, SEEK_END);
	fsiz = ftell(fp);
	rewind(fp);
	if (fsiz != (1024*32) && fsiz != (1024*24)) {
		fclose(fp);
		fprintf(stderr, "invalid rom file size\n");
		goto err_exit;
	}
	rsiz = fread(n80device.mem._rom, 1, fsiz, fp);
	fclose(fp);
	if (rsiz != fsiz) {
		fprintf(stderr, "invalid rom file read size\n");
		goto err_exit;
	}
	// PC-8031 rom file load
	BOOL fdu_rom_ready = FALSE;
	fp = fopen(n80lx.fdurom_filename, "rb");
	if (fp != NULL) {
		fseek(fp, 0, SEEK_END);
		fsiz = ftell(fp);
		rewind(fp);
		if (fsiz <= 8*1024) {
			rsiz = fread((LPVOID)n80device.fdu.mem._rom, 1, fsiz, fp);
			if (fsiz == rsiz)
				fdu_rom_ready = TRUE;
		}
		fclose(fp);
	}
	n80device.fdu.set_rom_ready(&n80device.fdu, fdu_rom_ready);

	// save cmt path
	char *buf = getcwd(n80lx.save_cmt_dir, 511-17);
	if (buf != NULL) {
		strcpy(n80lx.save_cmt_name, "cassette.cmt");
		sprintf(n80device.save_cmt_path, "%s/%s", n80lx.save_cmt_dir, n80lx.save_cmt_name);
	}
	printf("%s\n", n80device.save_cmt_path);
	#ifdef _ZAURUS_
	// open frame buffer
	n80lx.fbfd = open("/dev/fb", O_RDWR);
	if (!n80lx.fbfd) {
		fprintf(stderr, "can not open frambuffer!\n");
		return 1;
	}
	if (ioctl(n80lx.fbfd, FBIOGET_FSCREENINFO, &n80lx.finfo)) {
		fprintf(stderr, "can not purchase fb fix screen info\n");
		goto err_exit;
	}
	if (ioctl(n80lx.fbfd, FBIOGET_VSCREENINFO, &n80lx.vinfo)) {
		fprintf(stderr, "can not purchase fb variable screen info\n");
		goto err_exit;
	}
	// map frame buffer to memory
	len = n80lx.vinfo.xres * n80lx.vinfo.yres * n80lx.vinfo.bits_per_pixel / 8;
	n80lx.video = (LPUSHORT) mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, n80lx.fbfd, 0);
	if ((int)n80lx.video == -1) {
		fprintf(stderr, "can not map fb memory\n");
		goto err_exit;
	}
	clr_fb_display(n80lx.video);
	#else
	n80lx.video = (LPDWORD)aligned_alloc(128, 640*400*4);//  (LPDWORD) calloc(4, 640 * 400);
	if (n80lx.video != NULL) {
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTS) != 0) {
			fprintf(stderr, "can not init SDL2!\n");
			return 1;
		}
		n80lx.pipe_evt = SDL_RegisterEvents(1);
		if (n80lx.pipe_evt == ((Uint32)-1)) {
			fprintf(stderr, "can not register events\n");
			return 1;
		}
		n80lx.window = SDL_CreateWindow("N80 - PC-8001 Emulator",
			SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
//			640, 400, SDL_WINDOW_SHOWN | SDL_WINDOW_OPENGL);
			640, 400, SDL_WINDOW_SHOWN);
		if (n80lx.window == NULL) {
			fprintf(stderr, "can not open window!\n");
			goto err_exit;
		}
		n80lx.renderer = SDL_CreateRenderer(n80lx.window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
		if (n80lx.renderer == NULL) {
			SDL_DestroyWindow(n80lx.window);
			fprintf(stderr, "can not create renderer!\n");
			goto err_exit;
		}
		n80lx.texture = SDL_CreateTexture(n80lx.renderer,
			SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STREAMING, 640, 400);
		if (n80lx.texture == NULL) {
			SDL_DestroyRenderer(n80lx.renderer);
			SDL_DestroyWindow(n80lx.window);
			fprintf(stderr, "can not create texture!\n");
			goto err_exit;
		}
		SDL_ShowCursor(0);
	} else {
		fprintf(stderr, "can not allocate pixel buffer\n");
		goto err_exit;
	}
	#endif
	// alsa init
	if (init_alsa()) {
		fprintf(stderr, "can not open alsa\n");
		goto err_exit;
	}
	// recalc z80 thread interval according to alsa interval
	double alsa_interval_sec = (double)n80lx.period_frames / (double)n80lx.sampling_rate;
	double z80_interval_sec = alsa_interval_sec;
	n80device.calc_clk_sync(&n80device, n80device.param.hz, z80_interval_sec);
	fprintf(stdout, "alsa interval = %g(sec), z80 interval/ticks = %g(sec)/%d(T)\n", alsa_interval_sec, z80_interval_sec, n80device.param.sync_ticks);

	// create mutex
	pthread_mutex_init(&n80lx.vram_sync, NULL);
	pthread_mutex_init(&n80lx.pause_sync, NULL);

	// z80/alsa/disp threads start
	branch();

	// dispose mutex
	pthread_mutex_destroy(&n80lx.vram_sync);
	pthread_mutex_destroy(&n80lx.pause_sync);

	// free alsa
	dispose_alsa();

	#ifdef _ZAURUS_
	// free memory map
	munmap(n80lx.video, len);
	#else
	SDL_DestroyTexture(n80lx.texture);
	SDL_DestroyRenderer(n80lx.renderer);
	SDL_DestroyWindow(n80lx.window);
	#endif

	result = 0;
	
err_exit:
	// dispose bfd
	n80lx.bdf.dispose(&n80lx.bdf);
	// dispose json
	n80lx.ini.dispose(&n80lx.ini);
	// dispose N80 core
	n80device.dispose(&n80device);
	#ifdef _ZAURUS_
	close(n80lx.fbfd);
	#else
	free(n80lx.video);
	SDL_Quit();
	#endif
	return result;
}

int main(int argc, char *argv[]) {
	fprintf(stdout, "[\e[33;44;1mN80\e[m] PC-8001 Emulator 1.85 for Linux (SDL2)\n");
	fprintf(stdout, "Copyright 1993-2024 (c) H.Kanemoto.\n");
	for (int i = 0; i < argc; ++i) {
		if (!strcmp(argv[i], "-?") || !strcmp(argv[i], "--help")) {
			fprintf(stdout, "usage:n80\t[value | -? | --help]\n");
			printf ("\t\t-? or --help ..  this message.\n");
			printf ("\t\t[END] .........  quit from emulator.\n");
			printf ("\t\t[F7] ..........  system configuration.\n");
			printf ("\t\t[F9] ..........  CMT/N80 file chooser.\n");
			printf ("\t\t[F11] .........  toggle switch of HAL PCG-8100.\n");
			printf ("\t\t[F12] .........  reset button\n");
			printf ("\t\t[BS] ..........  stop key\n");
			return 0;
		}
	}
	fprintf(stdout, "show help when run with command line option \"-?\" or \"--help\".\n");
	return main_core();
}
