/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2017 H.Kanemoto
 **
 **		frame buffer operations
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#define	export
/**--------------------------------------------------------------------------
 **	external variables
 **------------------------------------------------------------------------*/
extern N80LX		n80lx;
#ifdef _DOSV_
typedef const void *LPCVOID;
extern void __cls_graphX(LPVOID vram, uint64_t lpitch);
extern void	__ovl_8020X(LPVOID txt, LPVOID pac, LPVOID font, LPVOID vram, LPCVOID palette, uint64_t lpitch, USHORT usepcg);
extern void	__ovl_8025X(LPVOID txt, LPVOID pac, LPVOID font, LPVOID vram, LPCVOID palette, uint64_t lpitch, USHORT usepcg);
extern void	__ovl_4020X(LPVOID txt, LPVOID pac, LPVOID font, LPVOID vram, LPCVOID palette, uint64_t lpitch, USHORT usepcg);
extern void	__ovl_4025X(LPVOID txt, LPVOID pac, LPVOID font, LPVOID vram, LPCVOID palette, uint64_t lpitch, USHORT usepcg);
#endif
/**--------------------------------------------------------------------------
 **	const variables
 **------------------------------------------------------------------------*/
static const DWORD	SPITZ_COLOR[8] = {
	0xff000000,
	0xff0000ff,
	0xff00ff00,
	0xff00ffff,
	0xffff0000,
	0xffff00ff,
	0xffffff00,
	0xffffffff
};
static const DWORD	SPITZ_COLOR2[8] = {
	0xff000000,
	0xff00007f,
	0xff007f00,
	0xff007f7f,
	0xff7f0000,
	0xff7f007f,
	0xff7f7f00,
	0xff7f7f7f
};
static const int	pitch = 640;
/**--------------------------------------------------------------------------
 **	internal variables/functions
 **------------------------------------------------------------------------*/
static void 		blt4020(int line);
static void 		blt4025(int line);
static void 		blt8020(int line);
static void 		blt8025(int line);
typedef void 		(*BLTFNC)(int);
static BLTFNC		bltfncs[] = {blt4020,blt4025,blt8020,blt8025};
static BLTFNC		fnc = blt4020;
static DWORD		cur_pix_rgb, cur_pix_rgb2;
static LPCDWORD		font_color2;
static int			vtbl_vram20[20], vtbl_vram25[25];
static DWORD		palette[2][8];
/**--------------------------------------------------------------------------
 **	N80 main thread
 **------------------------------------------------------------------------*/
export int init_font() {
	FILE *fp;
	int	i, len;
	
	// font file load
	fp = fopen("PC-8001.FON", "r");
	if (fp == NULL) {
		fprintf(stderr, "can not open font file\n");
		return 1;
	}
	len = fread(&n80device.font, 1, (1024*4), fp);
	fclose(fp);
	if (len != (1024*4)) {
		fprintf(stderr, "invalid font file size\n");
		return 1;
	}
	// フォントデータをPCGにコピー
	for (i = 0; i < 2048; ++i)
		n80device.pcg[i] = n80device.font[i];

	return 0;
}

export void enter(void) {
	pthread_mutex_lock(&n80lx.vram_sync);
}

export void leave(void) {
	pthread_mutex_unlock(&n80lx.vram_sync);
}

static void	chg_bltfunc(int mode) {
	fnc = bltfncs[mode];
}

static void chg_palette(BOOL interrace) {
	font_color2 = interrace ? SPITZ_COLOR2 : SPITZ_COLOR;
	for (int i = 0; i < 8; ++i)
		palette[1][i] = font_color2[i];
}
/**--------------------------------------------------------------------------
 **		set/reset pixel to vram buffer
 **------------------------------------------------------------------------*/
static inline void set_pixel_80(LPDWORD vram) {
	*vram           = cur_pix_rgb;
	*(vram+pitch)   = cur_pix_rgb2;
}
static inline void set_pixel_40(LPDWORD vram) {
	*vram           = cur_pix_rgb;
	*(vram+pitch)   = cur_pix_rgb2;
	*(vram+1)       = cur_pix_rgb;
	*(vram+pitch+1) = cur_pix_rgb2;
}
static inline void reset_pixel_80(LPDWORD vram) {
	*vram           = 0;
	*(vram+pitch)   = 0;
}
static inline void reset_pixel_40(LPDWORD vram) {
	*vram           = 0;
	*(vram+pitch)   = 0;
	*(vram+1)       = 0;
	*(vram+pitch+1) = 0;
}
/**--------------------------------------------------------------------------
 **		process font pattern by attribute
 **------------------------------------------------------------------------*/
static inline BYTE proc_video_pixcel25(BYTE attr, BYTE bit, int r) {
	if (attr & 0x80 && r == 7)	// under line
		bit |= 0xff;
	if (attr & 0x08 && r == 0)	// upper line
		bit |= 0xff;
	if (attr & 0x40)			// secret
		bit = 0;
	if (attr & 0x20)			// reverse
		bit = ~bit;
	return bit;
}
static inline BYTE proc_video_pixcel20(BYTE attr, BYTE bit, int r) {
	if (attr & 0x80 && r == 9)	// under line
		bit |= 0xff;
	if (attr & 0x08 && r == 0)	// upper line
		bit |= 0xff;
	if (attr & 0x40)			// secret
		bit = 0;
	if (attr & 0x20)			// reverse
		bit = ~bit;
	return bit;
}
/**--------------------------------------------------------------------------
 **		draw Text(40x20)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster4020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
	register LPDWORD _vram = (vram + (pitch * 2));
	register BYTE pat = proc_video_pixcel20(attr, fnt, r);

	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

	return _vram;
}
static void blt4020(int line) {
	if (line < 20) {
#ifdef _DOSV_
		LPDWORD	vaddr = (LPDWORD)n80lx.video;
		int		offset = vtbl_vram20[line];
		LPDWORD	vram = &vaddr[offset];

		__ovl_4020X(
			&n80device.attr._textcach[0],
			&n80device.attr._eval_attr[0],
			&n80device.font[0],
			vram,
			&palette[0][0],
			640*4,
			n80device.pcg_mode);
#else
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		int			col, offset = vtbl_vram20[line];

		for (col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			LPDWORD		vram  = &n80lx.video[offset + (col * 8)];
			int		 	color = attr & 0x07;

			cur_pix_rgb  = SPITZ_COLOR[color];
			cur_pix_rgb2 = font_color2[color];
			
			txts += 2;
			pacs += 2;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			vram = draw_font_raster4020(vram, 0, attr, 0);
			vram = draw_font_raster4020(vram, 1, attr, *fnt++);
			vram = draw_font_raster4020(vram, 2, attr, *fnt++);
			vram = draw_font_raster4020(vram, 3, attr, *fnt++);
			vram = draw_font_raster4020(vram, 4, attr, *fnt++);
			vram = draw_font_raster4020(vram, 5, attr, *fnt++);
			vram = draw_font_raster4020(vram, 6, attr, *fnt++);
			vram = draw_font_raster4020(vram, 7, attr, *fnt++);
			vram = draw_font_raster4020(vram, 8, attr, *fnt++);
			vram = draw_font_raster4020(vram, 9, attr, 0);
		}
#endif
	}
}
/**--------------------------------------------------------------------------
 **		draw Text(40x25)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster4025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
	register LPDWORD _vram = (vram + (pitch * 2));
	register BYTE pat = proc_video_pixcel25(attr, fnt, r);

	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

	return _vram;
}
static void blt4025(int line) {
	if (line < 25) {
#ifdef _DOSV_
		LPDWORD	vaddr = (LPDWORD)n80lx.video;
		int		offset = vtbl_vram25[line];
		LPDWORD	vram = &vaddr[offset];

		__ovl_4025X(
			&n80device.attr._textcach[0],
			&n80device.attr._eval_attr[0],
			&n80device.font[0],
			vram,
			&palette[0][0],
			640*4,
			n80device.pcg_mode);
#else
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		int			col, offset = vtbl_vram25[line];
		
		for (col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			LPDWORD		vram  = &n80lx.video[offset + (col * 8)];
			int		 	color = attr & 0x07;

			cur_pix_rgb  = SPITZ_COLOR[color];
			cur_pix_rgb2 = font_color2[color];
			
			txts += 2;
			pacs += 2;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			vram = draw_font_raster4025(vram, 0, attr, *fnt++);
			vram = draw_font_raster4025(vram, 1, attr, *fnt++);
			vram = draw_font_raster4025(vram, 2, attr, *fnt++);
			vram = draw_font_raster4025(vram, 3, attr, *fnt++);
			vram = draw_font_raster4025(vram, 4, attr, *fnt++);
			vram = draw_font_raster4025(vram, 5, attr, *fnt++);
			vram = draw_font_raster4025(vram, 6, attr, *fnt++);
			vram = draw_font_raster4025(vram, 7, attr, *fnt++);
		}
#endif
	}
}
/**--------------------------------------------------------------------------
 **		draw Text(80x20)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster8020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
	register LPDWORD _vram = (vram + (pitch * 2));
	register BYTE pat = proc_video_pixcel20(attr, fnt, r);

	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

	return _vram;
}
static void blt8020(int line) {
	if (line < 20) {
#ifdef _DOSV_
		LPDWORD	vaddr = (LPDWORD)n80lx.video;
		int		offset = vtbl_vram20[line];
		LPDWORD	vram = &vaddr[offset];

		__ovl_8020X(
			&n80device.attr._textcach[0],
			&n80device.attr._eval_attr[0],
			&n80device.font[0],
			vram,
			&palette[0][0],
			640*4,
			n80device.pcg_mode);
#else
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		int			col, offset = vtbl_vram20[line];
		
		for (col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			LPDWORD		vram  = &n80lx.video[offset + (col * 8)];
			int		 	color = attr & 0x07;

			cur_pix_rgb  = SPITZ_COLOR[color];
			cur_pix_rgb2 = font_color2[color];
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			vram = draw_font_raster8020(vram, 0, attr, 0);
			vram = draw_font_raster8020(vram, 1, attr, *fnt++);
			vram = draw_font_raster8020(vram, 2, attr, *fnt++);
			vram = draw_font_raster8020(vram, 3, attr, *fnt++);
			vram = draw_font_raster8020(vram, 4, attr, *fnt++);
			vram = draw_font_raster8020(vram, 5, attr, *fnt++);
			vram = draw_font_raster8020(vram, 6, attr, *fnt++);
			vram = draw_font_raster8020(vram, 7, attr, *fnt++);
			vram = draw_font_raster8020(vram, 8, attr, *fnt++);
			vram = draw_font_raster8020(vram, 9, attr, 0);
		}
#endif
	}
}
/**--------------------------------------------------------------------------
 **		draw Text(80x25)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster8025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
	register LPDWORD _vram = (vram + (pitch * 2));
	register BYTE pat = proc_video_pixcel25(attr, fnt, r);

	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

	return _vram;
}
static void blt8025(int line) {
	if (line < 25) {
#ifdef _DOSV_
		LPDWORD	vaddr = (LPDWORD)n80lx.video;
		int		offset = vtbl_vram25[line];
		LPDWORD	vram = &vaddr[offset];

		__ovl_8025X(
			&n80device.attr._textcach[0],
			&n80device.attr._eval_attr[0],
			&n80device.font[0],
			vram,
			&palette[0][0],
			640*4,
			n80device.pcg_mode);
#else
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		int			col, offset = vtbl_vram25[line];

		for (col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			LPDWORD		vram  = &n80lx.video[offset + (col * 8)];
			int		 	color = attr & 0x07;

			cur_pix_rgb  = SPITZ_COLOR[color];
			cur_pix_rgb2 = font_color2[color];
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			vram = draw_font_raster8025(vram, 0, attr, *fnt++);
			vram = draw_font_raster8025(vram, 1, attr, *fnt++);
			vram = draw_font_raster8025(vram, 2, attr, *fnt++);
			vram = draw_font_raster8025(vram, 3, attr, *fnt++);
			vram = draw_font_raster8025(vram, 4, attr, *fnt++);
			vram = draw_font_raster8025(vram, 5, attr, *fnt++);
			vram = draw_font_raster8025(vram, 6, attr, *fnt++);
			vram = draw_font_raster8025(vram, 7, attr, *fnt++);
		}
#endif
	}
}
/**--------------------------------------------------------------------------
 **		clear texture
 **------------------------------------------------------------------------*/
#ifndef _DOSV_
static const int vpitch[] = {
#ifdef _RASPBERRY_
	0,       pitch   ,pitch*2 ,pitch*3 ,pitch*4 ,pitch*5 ,pitch*6 ,pitch*7 ,pitch*8 ,pitch*9,
	pitch*10,pitch*11,pitch*12,pitch*13,pitch*14,pitch*15,pitch*16,pitch*17,pitch*18,pitch*19
#else
	0    ,640, 640*2, 640*3, 640*4, 640*5, 640*6, 640*7, 640*8, 640*9,
	640*10, 640*11, 640*12, 640*13, 640*14, 640*15, 640*16, 640*17, 640*18, 640*19
#endif
};
#endif
export void bltblank() {
#ifdef _DOSV_
	__cls_graphX(n80lx.video, 640*4);
#else
	int h = 0, i = 0;
	for (h = 0; h < 20; ++h) {
		int offset = vtbl_vram20[h];
		LPDWORD	vram = &n80lx.video[offset];
		for (i = 0; i < 640; ++i) {
			vram[i+vpitch[0]]  = 0;
			vram[i+vpitch[1]]  = 0;
			vram[i+vpitch[2]]  = 0;
			vram[i+vpitch[3]]  = 0;
			vram[i+vpitch[4]]  = 0;
			vram[i+vpitch[5]]  = 0;
			vram[i+vpitch[6]]  = 0;
			vram[i+vpitch[7]]  = 0;
			vram[i+vpitch[8]]  = 0;
			vram[i+vpitch[9]]  = 0;
			vram[i+vpitch[10]] = 0;
			vram[i+vpitch[11]] = 0;
			vram[i+vpitch[12]] = 0;
			vram[i+vpitch[13]] = 0;
			vram[i+vpitch[14]] = 0;
			vram[i+vpitch[15]] = 0;
			vram[i+vpitch[16]] = 0;
			vram[i+vpitch[17]] = 0;
			vram[i+vpitch[18]] = 0;
			vram[i+vpitch[19]] = 0;
		}
	}
#endif
}
/**--------------------------------------------------------------------------
 **		draw texture of screen
 **------------------------------------------------------------------------*/
static void reflesh_screen(void) {
	if ((n80device.crtc._text_visible & 0x06) == 0) {
		for (int i = 0; i < 25; ++i) {
			n80device.vram_sync_enter();
			n80device.attr.cache(i);
			n80device.vram_sync_leave();

			n80device.attr.eval(i);
			fnc(i);
		}
	} else {
		bltblank();
	}
}
/**--------------------------------------------------------------------------
 **		draw texture of line
 **------------------------------------------------------------------------*/
static inline void set_pixel_dbl(LPDWORD vram) {
	*vram = cur_pix_rgb;
}
static inline void reset_pixel_dbl(LPDWORD vram) {
	*vram = 0;
}
static inline BYTE proc_video_pixceldbl(BYTE attr, BYTE bit, int r) {
	if (attr & 0x80 && r == 15)	// under line
		bit |= 0xff;
	if (attr & 0x08 && r == 0)	// upper line
		bit |= 0xff;
	if (attr & 0x40)			// secret
		bit = 0;
	if (attr & 0x20)			// reverse
		bit = ~bit;
	return bit;
}
static inline LPDWORD draw_font_rasterdbl(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
	register LPDWORD _vram = vram + pitch;
	register BYTE pat = proc_video_pixceldbl(attr, fnt, r);

	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram++) : reset_pixel_dbl(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_dbl(vram) : reset_pixel_dbl(vram);

	return _vram;
}
static void blt8025k(int line) {
	if (line < 25) {
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		int			col, offset = vtbl_vram25[line];
		bool		dbl = false, zen = false;
		PBDFBMP		zenfnt = NULL;

		for (col = 0; col < 80; ++col) {
			BYTE		ascii = n80device.attr._textcach[col];
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			LPDWORD		vram  = &n80lx.video[offset + (col * 8)];
			int		 	color = attr & 0x07;

			cur_pix_rgb  = SPITZ_COLOR[color];
			cur_pix_rgb2 = col == 0 || col == 79 ? font_color2[color] : SPITZ_COLOR[color];
			
			if (attr & 0x10)				// graph
				fnt += 2048;

			if (zen && zenfnt != NULL) {
				fnt = &zenfnt->bitmap[1][0];
				zenfnt = NULL;
				zen    = false;
				dbl    = true;
			} else if ((ascii >= 0x80 && ascii <= 0xa0) || (ascii >= 0xe0 && ascii <= 0xff)) {
				if (col < 79) {
					BYTE ascii2 = n80device.attr._textcach[col + 1];
					uint16_t sjis = ascii;
					sjis <<= 8;
					sjis |= ascii2;
					zenfnt = n80lx.bdf.find(&n80lx.bdf, sjis);
					if (zenfnt != NULL) {
						fnt = &zenfnt->bitmap[0][0];
						zen = true;
						dbl = true;
					}
				}
			}
			if (dbl) {
				dbl = false;
				vram = draw_font_rasterdbl(vram, 0, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 1, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 2, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 3, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 4, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 5, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 6, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 7, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 8, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 9, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 10, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 11, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 12, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 13, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 14, attr, *fnt++);
				vram = draw_font_rasterdbl(vram, 15, attr, *fnt++);
			} else {
				vram = draw_font_raster8025(vram, 0, attr, *fnt++);
				vram = draw_font_raster8025(vram, 1, attr, *fnt++);
				vram = draw_font_raster8025(vram, 2, attr, *fnt++);
				vram = draw_font_raster8025(vram, 3, attr, *fnt++);
				vram = draw_font_raster8025(vram, 4, attr, *fnt++);
				vram = draw_font_raster8025(vram, 5, attr, *fnt++);
				vram = draw_font_raster8025(vram, 6, attr, *fnt++);
				vram = draw_font_raster8025(vram, 7, attr, *fnt++);
			}
		}
	}
}
export void draw_text(int line, LPBYTE attr, LPBYTE text) {
	memcpy(n80device.attr._textcach, text, 80);
	memcpy(n80device.attr._eval_attr, attr, 80);
	blt8025(line);
}
export void draw_textk(int line, LPBYTE attr, LPBYTE text) {
	memcpy(n80device.attr._textcach, text, 80);
	memcpy(n80device.attr._eval_attr, attr, 80);
	blt8025k(line);
}
export void _n80_init_blt(void) {
	n80device.blt.chg_crt_mode	= chg_bltfunc;
	n80device.blt.reflesh		= reflesh_screen;
	n80device.blt.chg_palette	= chg_palette;

	for (int i = 0; i < 8; ++i)
		palette[0][i] = SPITZ_COLOR[i];
	chg_palette(n80device.param.interlace);
	for (int line = 0; line < 20; ++line)
		vtbl_vram20[line] = line * 10 * 2 * pitch;
	for (int line = 0; line < 25; ++line)
		vtbl_vram25[line] = line * 8 * 2 * pitch;
}
